import numpy as np
from scipy.spatial import cKDTree, distance

from nellie import xp, ndi, logger, device_type
from nellie.im_info.im_info import ImInfo
from nellie.utils.general import get_reshaped_image


class Markers:
    def __init__(self, im_info: ImInfo, num_t=None,
                 min_radius_um=0.20, max_radius_um=1, use_im='distance', num_sigma=5):
        self.im_info = im_info
        self.num_t = num_t
        if self.im_info.no_t:
            self.num_t = 1
        elif num_t is None:  # and not self.im_info.no_t:
            self.num_t = im_info.shape[im_info.axes.index('T')]
        if not self.im_info.no_z:
            self.z_ratio = self.im_info.dim_sizes['Z'] / self.im_info.dim_sizes['X']
        self.min_radius_um = max(min_radius_um, self.im_info.dim_sizes['X'])
        self.max_radius_um = max_radius_um

        self.min_radius_px = self.min_radius_um / self.im_info.dim_sizes['X']
        self.max_radius_px = self.max_radius_um / self.im_info.dim_sizes['X']
        self.use_im = use_im
        self.num_sigma = num_sigma

        self.shape = ()

        self.im_memmap = None
        self.im_frangi_memmap = None
        self.label_memmap = None
        self.im_marker_memmap = None
        self.im_distance_memmap = None
        self.im_border_memmap = None

        self.debug = None

    def _get_sigma_vec(self, sigma):
        if self.im_info.no_z:
            sigma_vec = (sigma, sigma)
        else:
            sigma_vec = (sigma / self.z_ratio, sigma, sigma)
        return sigma_vec

    def _set_default_sigmas(self):
        logger.debug('Setting sigma values.')
        min_sigma_step_size = 0.2

        self.sigma_min = self.min_radius_px / 2
        self.sigma_max = self.max_radius_px / 3

        sigma_step_size_calculated = (self.sigma_max - self.sigma_min) / self.num_sigma
        sigma_step_size = max(min_sigma_step_size, sigma_step_size_calculated)  # Avoid taking too small of steps.

        self.sigmas = list(xp.arange(self.sigma_min, self.sigma_max, sigma_step_size))
        logger.debug(f'Calculated sigma step size = {sigma_step_size_calculated}. Sigmas = {self.sigmas}')

    def _get_t(self):
        if self.num_t is None:
            if self.im_info.no_t:
                self.num_t = 1
            else:
                self.num_t = self.im_info.shape[self.im_info.axes.index('T')]
        else:
            return

    def _allocate_memory(self):
        logger.debug('Allocating memory for mocap marking.')

        label_memmap = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_instance_label'])
        self.label_memmap = get_reshaped_image(label_memmap, self.num_t, self.im_info)

        im_memmap = self.im_info.get_im_memmap(self.im_info.im_path)
        self.im_memmap = get_reshaped_image(im_memmap, self.num_t, self.im_info)

        im_frangi_memmap = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_frangi'])
        self.im_frangi_memmap = get_reshaped_image(im_frangi_memmap, self.num_t, self.im_info)
        self.shape = self.label_memmap.shape

        im_marker_path = self.im_info.pipeline_paths['im_marker']
        self.im_marker_memmap = self.im_info.allocate_memory(im_marker_path, shape=self.shape,
                                                             dtype='uint8',
                                                             description='mocap marker image',
                                                             return_memmap=True)

        im_distance_path = self.im_info.pipeline_paths['im_distance']
        self.im_distance_memmap = self.im_info.allocate_memory(im_distance_path, shape=self.shape,
                                                               dtype='float',
                                                               description='distance transform image',
                                                               return_memmap=True)

        im_border_path = self.im_info.pipeline_paths['im_border']
        self.im_border_memmap = self.im_info.allocate_memory(im_border_path, shape=self.shape,
                                                             dtype='uint8',
                                                             description='border image',
                                                             return_memmap=True)

    def _distance_im(self, mask):
        border_mask = ndi.binary_dilation(mask, iterations=1) ^ mask

        if device_type == 'cuda':
            mask_coords = xp.argwhere(mask).get()
            border_mask_coords = xp.argwhere(border_mask).get()
        else:
            mask_coords = xp.argwhere(mask)
            border_mask_coords = xp.argwhere(border_mask)

        border_tree = cKDTree(border_mask_coords)
        dist, _ = border_tree.query(mask_coords, k=1, distance_upper_bound=self.max_radius_px * 2)
        distances_im_frame = xp.zeros_like(mask, dtype='float32')
        if self.im_info.no_z:
            distances_im_frame[mask_coords[:, 0], mask_coords[:, 1]] = dist
        else:
            distances_im_frame[mask_coords[:, 0], mask_coords[:, 1], mask_coords[:, 2]] = dist
        # any inf pixels get set to upper bound
        distances_im_frame[distances_im_frame == xp.inf] = self.max_radius_px * 2
        return distances_im_frame, border_mask

    def _remove_close_peaks(self, coord, check_im):
        check_im_max = ndi.maximum_filter(check_im, size=3, mode='nearest')
        if not self.im_info.no_z:
            intensities = check_im_max[coord[:, 0], coord[:, 1], coord[:, 2]]
        else:
            intensities = check_im_max[coord[:, 0], coord[:, 1]]

        # sort to remove peaks that are too close by keeping the brightest peak
        idx_maxsort = np.argsort(-intensities)

        if device_type == 'cuda':
            coord_sorted = coord[idx_maxsort].get()
        else:
            coord_sorted = coord[idx_maxsort]

        tree = cKDTree(coord_sorted)
        min_dist = 2
        indices = tree.query_ball_point(coord_sorted, r=min_dist, p=2, workers=-1)
        rejected_peaks_indices = set()
        naccepted = 0
        for idx, candidates in enumerate(indices):
            if idx not in rejected_peaks_indices:
                # keep current point and the points at exactly spacing from it
                candidates.remove(idx)
                dist = distance.cdist([coord_sorted[idx]],
                                      coord_sorted[candidates],
                                      distance.minkowski,
                                      p=2).reshape(-1)
                candidates = [c for c, d in zip(candidates, dist)
                              if d < min_dist]

                rejected_peaks_indices.update(candidates)
                naccepted += 1

        cleaned_coords = np.delete(coord_sorted, tuple(rejected_peaks_indices), axis=0)

        return cleaned_coords

    def _local_max_peak(self, use_im, mask, distance_im):
        lapofg = xp.empty(((len(self.sigmas),) + use_im.shape), dtype=float)
        for i, s in enumerate(self.sigmas):
            sigma_vec = self._get_sigma_vec(s)
            current_lapofg = -ndi.gaussian_laplace(use_im, sigma_vec) * xp.mean(s) ** 2
            current_lapofg[current_lapofg < 0] = 0
            lapofg[i] = current_lapofg

        filt_footprint = xp.ones((3,) * (use_im.ndim + 1))
        max_filt = ndi.maximum_filter(lapofg, footprint=filt_footprint, mode='nearest')
        peaks = xp.empty(lapofg.shape, dtype=bool)
        for filt_slice, max_filt_slice in enumerate(max_filt):
            peaks[filt_slice] = (xp.asarray(lapofg[filt_slice]) == xp.asarray(max_filt_slice))  # * max_filt_mask
        distance_mask = distance_im > 0
        peaks = peaks * mask * distance_mask
        # get the coordinates of all true pixels in peaks
        coords = xp.max(peaks, axis=0)
        coords_idx = xp.argwhere(coords)
        return coords_idx

    def _run_frame(self, t):
        logger.info(f'Running motion capture marking, volume {t}/{self.num_t - 1}')
        intensity_frame = xp.asarray(self.im_memmap[t])
        mask_frame = xp.asarray(self.label_memmap[t] > 0)
        distance_im, border_mask = self._distance_im(mask_frame)
        if self.use_im == 'distance':
            peak_coords = self._local_max_peak(distance_im, mask_frame, distance_im)
        elif self.use_im == 'frangi':
            peak_coords = self._local_max_peak(xp.asarray(self.im_frangi_memmap[t]), mask_frame, distance_im)
        peak_coords = self._remove_close_peaks(peak_coords, intensity_frame)
        peak_im = xp.zeros_like(mask_frame)
        peak_im[tuple(peak_coords.T)] = 1
        if device_type == "cuda":
            return peak_im.get(), distance_im.get(), border_mask.get()
        else:
            return peak_im, distance_im, border_mask

    def _run_mocap_marking(self):
        for t in range(self.num_t):
            marker_frame = self._run_frame(t)
            if self.im_marker_memmap.shape != self.shape and self.im_info.no_t:
                self.im_marker_memmap[:], self.im_distance_memmap[:], self.im_border_memmap[:] = marker_frame
            else:
                self.im_marker_memmap[t], self.im_distance_memmap[t], self.im_border_memmap[t] = marker_frame
            self.im_marker_memmap.flush()
            self.im_distance_memmap.flush()
            self.im_border_memmap.flush()

    def run(self):
        # if self.im_info.no_t:
        #     return
        self._get_t()
        self._allocate_memory()
        self._set_default_sigmas()
        self._run_mocap_marking()


if __name__ == "__main__":
    im_path = r"D:\test_files\nelly_smorgasbord\deskewed-iono_pre.ome.tif"
    im_info = ImInfo(im_path)
    num_t = 3
    markers = Markers(im_info, num_t=num_t)
    markers.run()



import numpy as np
from scipy.spatial.distance import cdist

from nellie import xp, ndi, logger
from nellie.im_info.im_info import ImInfo
from nellie.utils.general import get_reshaped_image


class HuMomentTracking:
    def __init__(self, im_info: ImInfo, num_t=None,
                 max_distance_um=1):
        self.im_info = im_info
        self.num_t = num_t
        if num_t is None and not self.im_info.no_t:
            self.num_t = im_info.shape[im_info.axes.index('T')]

        if self.im_info.no_z:
            self.scaling = (im_info.dim_sizes['Y'], im_info.dim_sizes['X'])
        else:
            self.scaling = (im_info.dim_sizes['Z'], im_info.dim_sizes['Y'], im_info.dim_sizes['X'])

        self.max_distance_um = max_distance_um

        self.vector_start_coords = []
        self.vectors = []
        self.vector_magnitudes = []

        self.shape = ()

        self.im_memmap = None
        self.im_frangi_memmap = None
        self.im_distance_memmap = None
        self.im_marker_memmap = None
        self.flow_vector_array_path = None

        self.debug = None

    def _calculate_normalized_moments(self, images):
        # I know the broadcasting is super confusing, but it makes it so much faster (400x)...

        num_images, height, width = images.shape
        extended_images = images[:, :, :, None, None]  # shape (num_images, height, width, 1, 1)

        # pre-compute meshgrid
        x, y = xp.meshgrid(xp.arange(width), xp.arange(height))

        # reshape for broadcasting
        x = x[None, :, :, None, None]
        y = y[None, :, :, None, None]

        # raw moments
        M = xp.sum(extended_images * (x ** xp.arange(4)[None, None, None, :, None]) *
                   (y ** xp.arange(4)[None, None, None, None, :]), axis=(1, 2))

        # central Moments; compute x_bar and y_bar
        x_bar = M[:, 1, 0] / M[:, 0, 0]
        y_bar = M[:, 0, 1] / M[:, 0, 0]

        x_bar = x_bar[:, None, None, None, None]
        y_bar = y_bar[:, None, None, None, None]

        # calculate mu using broadcasting
        mu = xp.sum(extended_images * (x - x_bar) ** xp.arange(4)[None, None, None, :, None] *
                    (y - y_bar) ** xp.arange(4)[None, None, None, None, :], axis=(1, 2))

        # normalized moments
        i_plus_j = xp.arange(4)[:, None] + xp.arange(4)[None, :]
        eta = mu / (M[:, 0, 0][:, None, None] ** ((i_plus_j[None, :, :] + 2) / 2))

        return eta

    def _calculate_hu_moments(self, eta):
        num_images = eta.shape[0]
        hu = xp.zeros((num_images, 6))  # initialize Hu moments for each image

        hu[:, 0] = eta[:, 2, 0] + eta[:, 0, 2]
        hu[:, 1] = (eta[:, 2, 0] - eta[:, 0, 2]) ** 2 + 4 * eta[:, 1, 1] ** 2
        hu[:, 2] = (eta[:, 3, 0] - 3 * eta[:, 1, 2]) ** 2 + (3 * eta[:, 2, 1] - eta[:, 0, 3]) ** 2
        hu[:, 3] = (eta[:, 3, 0] + eta[:, 1, 2]) ** 2 + (eta[:, 2, 1] + eta[:, 0, 3]) ** 2
        hu[:, 4] = (eta[:, 3, 0] - 3 * eta[:, 1, 2]) * (eta[:, 3, 0] + eta[:, 1, 2]) * \
                   ((eta[:, 3, 0] + eta[:, 1, 2]) ** 2 - 3 * (eta[:, 2, 1] + eta[:, 0, 3]) ** 2) + \
                   (3 * eta[:, 2, 1] - eta[:, 0, 3]) * (eta[:, 2, 1] + eta[:, 0, 3]) * \
                   (3 * (eta[:, 3, 0] + eta[:, 1, 2]) ** 2 - (eta[:, 2, 1] + eta[:, 0, 3]) ** 2)
        hu[:, 5] = (eta[:, 2, 0] - eta[:, 0, 2]) * \
                   ((eta[:, 3, 0] + eta[:, 1, 2]) ** 2 - (eta[:, 2, 1] + eta[:, 0, 3]) ** 2) + \
                   4 * eta[:, 1, 1] * (eta[:, 3, 0] + eta[:, 1, 2]) * (eta[:, 2, 1] + eta[:, 0, 3])
        # don't want mirror symmetry invariance.. doesn't make sense for our application
        # hu[:, 6] = (3 * eta[:, 2, 1] - eta[:, 0, 3]) * (eta[:, 3, 0] + eta[:, 1, 2]) * \
        #            ((eta[:, 3, 0] + eta[:, 1, 2]) ** 2 - 3 * (eta[:, 2, 1] + eta[:, 0, 3]) ** 2) - \
        #            (eta[:, 3, 0] - 3 * eta[:, 1, 2]) * (eta[:, 2, 1] + eta[:, 0, 3]) * \
        #            (3 * (eta[:, 3, 0] + eta[:, 1, 2]) ** 2 - (eta[:, 2, 1] + eta[:, 0, 3]) ** 2)

        return hu  # return the first 5 Hu moments for each image

    def _calculate_mean_and_variance(self, images):
        num_images = images.shape[0]
        features = xp.zeros((num_images, 2))
        mask = images != 0

        if self.im_info.no_z:
            axis = (1, 2)
        else:
            axis = (1, 2, 3)
        count_nonzero = xp.sum(mask, axis=axis)
        sum_nonzero = xp.sum(images * mask, axis=axis)
        sumsq_nonzero = xp.sum((images * mask) ** 2, axis=axis)

        mean = sum_nonzero / count_nonzero
        variance = (sumsq_nonzero - (sum_nonzero ** 2) / count_nonzero) / count_nonzero

        features[:, 0] = mean
        features[:, 1] = variance
        return features

    def _get_im_bounds(self, markers, distance_frame):
        if not self.im_info.no_z:
            radii = distance_frame[markers[:, 0], markers[:, 1], markers[:, 2]]
        else:
            radii = distance_frame[markers[:, 0], markers[:, 1]]
        marker_radii = xp.ceil(radii)
        low_0 = xp.clip(markers[:, 0] - marker_radii, 0, self.shape[1])
        high_0 = xp.clip(markers[:, 0] + (marker_radii + 1), 0, self.shape[1])
        low_1 = xp.clip(markers[:, 1] - marker_radii, 0, self.shape[2])
        high_1 = xp.clip(markers[:, 1] + (marker_radii + 1), 0, self.shape[2])
        if not self.im_info.no_z:
            low_2 = xp.clip(markers[:, 2] - marker_radii, 0, self.shape[3])
            high_2 = xp.clip(markers[:, 2] + (marker_radii + 1), 0, self.shape[3])
            return low_0, high_0, low_1, high_1, low_2, high_2
        return low_0, high_0, low_1, high_1

    def _get_sub_volumes(self, im_frame, im_bounds, max_radius):
        if self.im_info.no_z:
            y_low, y_high, x_low, x_high = im_bounds
        else:
            z_low, z_high, y_low, y_high, x_low, x_high = im_bounds

        # preallocate arrays
        if self.im_info.no_z:
            sub_volumes = xp.zeros((len(y_low), max_radius, max_radius))
        else:
            sub_volumes = xp.zeros((len(y_low), max_radius, max_radius, max_radius))

        # extract sub-volumes
        for i in range(len(y_low)):
            if self.im_info.no_z:
                yl, yh, xl, xh = int(y_low[i]), int(y_high[i]), int(x_low[i]), int(x_high[i])
                sub_volumes[i, :yh - yl, :xh - xl] = im_frame[yl:yh, xl:xh]
            else:
                zl, zh, yl, yh, xl, xh = int(z_low[i]), int(z_high[i]), int(y_low[i]), int(y_high[i]), int(
                    x_low[i]), int(x_high[i])
                sub_volumes[i, :zh - zl, :yh - yl, :xh - xl] = im_frame[zl:zh, yl:yh, xl:xh]

        return sub_volumes

    def _get_orthogonal_projections(self, sub_volumes):
        # max projections along each axis
        z_projections = xp.max(sub_volumes, axis=1)
        y_projections = xp.max(sub_volumes, axis=2)
        x_projections = xp.max(sub_volumes, axis=3)

        return z_projections, y_projections, x_projections

    def _get_t(self):
        if self.num_t is None:
            if self.im_info.no_t:
                self.num_t = 1
            else:
                self.num_t = self.im_info.shape[self.im_info.axes.index('T')]
        else:
            return

    def _allocate_memory(self):
        logger.debug('Allocating memory for hu-based tracking.')
        label_memmap = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_instance_label'])
        self.label_memmap = get_reshaped_image(label_memmap, self.num_t, self.im_info)

        im_memmap = self.im_info.get_im_memmap(self.im_info.im_path)
        self.im_memmap = get_reshaped_image(im_memmap, self.num_t, self.im_info)

        im_frangi_memmap = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_frangi'])
        self.im_frangi_memmap = get_reshaped_image(im_frangi_memmap, self.num_t, self.im_info)
        self.shape = self.label_memmap.shape

        im_marker_memmap = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_marker'])
        self.im_marker_memmap = get_reshaped_image(im_marker_memmap, self.num_t, self.im_info)

        im_distance_memmap = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_distance'])
        self.im_distance_memmap = get_reshaped_image(im_distance_memmap, self.num_t, self.im_info)

        self.flow_vector_array_path = self.im_info.pipeline_paths['flow_vector_array']

    def _get_hu_moments(self, sub_volumes):
        if self.im_info.no_z:
            etas = self._calculate_normalized_moments(sub_volumes)
            hu_moments = self._calculate_hu_moments(etas)
            return hu_moments
        intensity_projections = self._get_orthogonal_projections(sub_volumes)
        etas_z = self._calculate_normalized_moments(intensity_projections[0])
        etas_y = self._calculate_normalized_moments(intensity_projections[1])
        etas_x = self._calculate_normalized_moments(intensity_projections[2])
        hu_moments_z = self._calculate_hu_moments(etas_z)
        hu_moments_y = self._calculate_hu_moments(etas_y)
        hu_moments_x = self._calculate_hu_moments(etas_x)
        hu_moments = xp.concatenate((hu_moments_z, hu_moments_y, hu_moments_x), axis=1)
        return hu_moments

    def _concatenate_hu_matrices(self, hu_matrices):
        return xp.concatenate(hu_matrices, axis=1)

    def _get_feature_matrix(self, t):
        intensity_frame = xp.array(self.im_memmap[t])
        frangi_frame = xp.array(self.im_frangi_memmap[t])
        frangi_frame[frangi_frame > 0] = xp.log10(frangi_frame[frangi_frame > 0])
        frangi_frame[frangi_frame < 0] -= xp.min(frangi_frame[frangi_frame < 0])

        distance_frame = xp.array(self.im_distance_memmap[t])
        distance_max_frame = ndi.maximum_filter(distance_frame, size=3) * 2

        marker_frame = xp.array(self.im_marker_memmap[t]) > 0
        marker_indices = xp.argwhere(marker_frame)

        region_bounds = self._get_im_bounds(marker_indices, distance_max_frame)
        max_radius = int(xp.ceil(xp.max(distance_max_frame[marker_frame]))) * 2 + 1

        intensity_sub_volumes = self._get_sub_volumes(intensity_frame, region_bounds, max_radius)
        frangi_sub_volumes = self._get_sub_volumes(frangi_frame, region_bounds, max_radius)

        intensity_stats = self._calculate_mean_and_variance(intensity_sub_volumes)
        frangi_stats = self._calculate_mean_and_variance(frangi_sub_volumes)
        stats_feature_matrix = self._concatenate_hu_matrices([intensity_stats, frangi_stats])

        intensity_hus = self._get_hu_moments(intensity_sub_volumes)
        log_hu_feature_matrix = -1 * xp.copysign(1.0, intensity_hus) * xp.log10(xp.abs(intensity_hus))
        log_hu_feature_matrix[xp.isinf(log_hu_feature_matrix)] = xp.nan

        return stats_feature_matrix, log_hu_feature_matrix

    def _get_distance_mask(self, t):
        marker_frame_pre = np.array(self.im_marker_memmap[t - 1]) > 0
        marker_indices_pre = np.argwhere(marker_frame_pre)
        marker_indices_pre_scaled = marker_indices_pre * self.scaling
        marker_frame_post = np.array(self.im_marker_memmap[t]) > 0
        marker_indices_post = np.argwhere(marker_frame_post)
        marker_indices_post_scaled = marker_indices_post * self.scaling

        distance_matrix = cdist(marker_indices_post_scaled, marker_indices_pre_scaled)
        distance_mask = xp.array(distance_matrix) < self.max_distance_um
        distance_matrix = distance_matrix / self.max_distance_um  # normalize to furthest possible distance
        return distance_matrix, distance_mask

    def _get_difference_matrix(self, m1, m2):
        m1_reshaped = m1[:, xp.newaxis, :].astype(xp.float16)
        m2_reshaped = m2[xp.newaxis, :, :].astype(xp.float16)
        difference_matrix = xp.abs(m1_reshaped - m2_reshaped)
        return difference_matrix

    def _zscore_normalize(self, m, mask):
        depth = m.shape[2]

        sum_mask = xp.sum(mask)
        mean_vals = xp.zeros(depth)
        std_vals = xp.zeros(depth)

        # calculate mean values slice by slice
        for d in range(depth):
            slice_m = m[:, :, d]
            mean_vals[d] = xp.sum(slice_m * mask) / sum_mask

        # calculate std values slice by slice
        for d in range(depth):
            slice_m = m[:, :, d]
            std_vals[d] = xp.sqrt(xp.sum((slice_m - mean_vals[d]) ** 2 * mask) / sum_mask)

        # normalize and set to infinity where mask is 0
        for d in range(depth):
            slice_m = m[:, :, d]
            slice_m -= mean_vals[d]
            slice_m /= std_vals[d]
            slice_m[mask == 0] = xp.inf

        return m

    def _get_cost_matrix(self, t, stats_vecs, pre_stats_vecs, hu_vecs, pre_hu_vecs):
        distance_matrix, distance_mask = self._get_distance_mask(t)
        z_score_distance_matrix = self._zscore_normalize(xp.array(distance_matrix)[..., xp.newaxis],
                                                         distance_mask).astype(xp.float16)
        del distance_matrix
        stats_matrix = self._get_difference_matrix(stats_vecs, pre_stats_vecs)
        z_score_stats_matrix = (self._zscore_normalize(stats_matrix, distance_mask) / stats_matrix.shape[2]).astype(
            xp.float16)
        del stats_matrix
        hu_matrix = self._get_difference_matrix(hu_vecs, pre_hu_vecs)
        z_score_hu_matrix = (self._zscore_normalize(hu_matrix, distance_mask) / hu_matrix.shape[2]).astype(xp.float16)
        del hu_matrix, distance_mask
        z_score_matrix = xp.concatenate((z_score_distance_matrix, z_score_stats_matrix, z_score_hu_matrix),
                                        axis=2).astype(xp.float16)
        cost_matrix = xp.nansum(z_score_matrix, axis=2).astype(xp.float16)

        return cost_matrix

    def _find_best_matches(self, cost_matrix):
        candidates = []
        cost_cutoff = 1

        # find row-wise minimums
        row_min_idx = xp.argmin(cost_matrix, axis=1)
        row_min_val = xp.min(cost_matrix, axis=1)

        # find column-wise minimums
        col_min_idx = xp.argmin(cost_matrix, axis=0)
        col_min_val = xp.min(cost_matrix, axis=0)

        row_matches = []
        col_matches = []
        costs = []

        # store each row's and column's minimums as candidates for matching
        for i, (r_idx, r_val) in enumerate(zip(row_min_idx, row_min_val)):
            if r_val > cost_cutoff:
                continue
            candidates.append((int(i), int(r_idx), float(r_val)))
            row_matches.append(int(i))
            col_matches.append(int(r_idx))
            costs.append(float(r_val))

        for j, (c_idx, c_val) in enumerate(zip(col_min_idx, col_min_val)):
            if c_val > cost_cutoff:
                continue
            candidates.append((int(c_idx), int(j), float(c_val)))
            row_matches.append(int(c_idx))
            col_matches.append(int(j))
            costs.append(float(c_val))

        return row_matches, col_matches, costs

    def _run_hu_tracking(self):
        pre_stats_vecs = None
        pre_hu_vecs = None
        flow_vector_array = None
        for t in range(self.num_t):
            logger.debug(f'Running hu-moment tracking for frame {t + 1} of {self.num_t}')
            stats_vecs, hu_vecs = self._get_feature_matrix(t)
            # todo make distance weighting be dependent on number of seconds between frames (more uncertain with more time)
            #  could also vary with size (radius) based on diffusion coefficient. bigger = probably closer
            if pre_stats_vecs is None or pre_hu_vecs is None:
                pre_stats_vecs = stats_vecs
                pre_hu_vecs = hu_vecs
                continue
            cost_matrix = self._get_cost_matrix(t, stats_vecs, pre_stats_vecs, hu_vecs, pre_hu_vecs)
            row_indices, col_indices, costs = self._find_best_matches(cost_matrix)
            pre_marker_indices = np.argwhere(self.im_marker_memmap[t - 1])[col_indices]
            marker_indices = np.argwhere(self.im_marker_memmap[t])[row_indices]
            vecs = np.array(marker_indices) - np.array(pre_marker_indices)

            pre_stats_vecs = stats_vecs
            pre_hu_vecs = hu_vecs

            costs = np.array(costs)
            if self.im_info.no_z:
                idx0_y, idx0_x = pre_marker_indices.T
                vec_y, vec_x = vecs.T
                frame_vector_array = np.concatenate((np.array([t - 1] * len(vec_y))[:, np.newaxis],
                                                     idx0_y[:, np.newaxis], idx0_x[:, np.newaxis],
                                                     vec_y[:, np.newaxis], vec_x[:, np.newaxis],
                                                     costs[:, np.newaxis]), axis=1)
            else:
                idx0_z, idx0_y, idx0_x = pre_marker_indices.T
                vec_z, vec_y, vec_x = vecs.T
                frame_vector_array = np.concatenate((np.array([t - 1] * len(vec_z))[:, np.newaxis],
                                                     idx0_z[:, np.newaxis], idx0_y[:, np.newaxis],
                                                     idx0_x[:, np.newaxis],
                                                     vec_z[:, np.newaxis], vec_y[:, np.newaxis], vec_x[:, np.newaxis],
                                                     costs[:, np.newaxis]), axis=1)
            if flow_vector_array is None:
                flow_vector_array = frame_vector_array
            else:
                flow_vector_array = np.concatenate((flow_vector_array, frame_vector_array), axis=0)
            del frame_vector_array

        # save the array
        np.save(self.flow_vector_array_path, flow_vector_array)

    def run(self):
        if self.im_info.no_t:
            return
        self._get_t()
        self._allocate_memory()
        self._run_hu_tracking()


if __name__ == "__main__":
    im_path = r"D:\test_files\nelly_smorgasbord\deskewed-iono_pre.ome.tif"
    im_info = ImInfo(im_path)
    hu = HuMomentTracking(im_info, num_t=2)
    hu.run()



from nellie.feature_extraction.hierarchical import Hierarchy
from nellie.im_info.im_info import ImInfo
from nellie.segmentation.filtering import Filter
from nellie.segmentation.labelling import Label
from nellie.segmentation.mocap_marking import Markers
from nellie.segmentation.networking import Network
from nellie.tracking.hu_tracking import HuMomentTracking
from nellie.tracking.voxel_reassignment import VoxelReassigner


def run(im_path, num_t=None, remove_edges=True, ch=0):
    im_info = ImInfo(im_path, ch=ch)

    preprocessing = Filter(im_info, num_t, remove_edges=remove_edges)
    preprocessing.run()

    segmenting = Label(im_info, num_t)
    segmenting.run()

    networking = Network(im_info, num_t)
    networking.run()

    mocap_marking = Markers(im_info, num_t)
    mocap_marking.run()

    hu_tracking = HuMomentTracking(im_info, num_t)
    hu_tracking.run()

    vox_reassign = VoxelReassigner(im_info, num_t)
    vox_reassign.run()

    hierarchy = Hierarchy(im_info, num_t)
    hierarchy.run()

    return im_info


if __name__ == "__main__":
    # Single file run
    im_path = r"/Users/austin/Downloads/test.tif"
    im_info = run(im_path, remove_edges=False, ch=0)

    # # Directory bactch run
    # import os
    # top_dir = r"D:\test_files\nellie_longer_smorgasbord"
    # ch = 0
    # num_t = None
    # # get all non-folder files
    # all_files = os.listdir(top_dir)
    # all_files = [os.path.join(top_dir, file) for file in all_files if not os.path.isdir(os.path.join(top_dir, file))]
    # for file_num, tif_file in enumerate(all_files):
    #     # for ch in range(1):
    #     print(f'Processing file {file_num + 1} of {len(all_files)}, channel {ch + 1} of 1')
    #     im_info = run(tif_file, remove_edges=True, ch=ch, num_t=num_t)



import os

from napari.utils.notifications import show_info
from qtpy.QtWidgets import QWidget, QPushButton, QLabel, QGridLayout, QSpinBox, QCheckBox
from nellie import logger
from nellie.feature_extraction.hierarchical import Hierarchy
from nellie.segmentation.filtering import Filter
from nellie.segmentation.labelling import Label
from nellie.segmentation.mocap_marking import Markers
from nellie.segmentation.networking import Network
from nellie.tracking.hu_tracking import HuMomentTracking
from nellie.tracking.voxel_reassignment import VoxelReassigner
from nellie.utils.general import get_reshaped_image


class NellieProcessor(QWidget):
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', nellie, parent=None):
        super().__init__(parent)
        self.nellie = nellie
        self.viewer = napari_viewer

        # Label above the spinner box
        self.channel_label = QLabel("Channel to analyze:")

        self.channel_input = QSpinBox()
        self.channel_input.setRange(0, 0)
        self.channel_input.setValue(0)
        self.channel_input.setEnabled(False)
        self.channel_input.valueChanged.connect(self.change_channel)

        # Label above the spinner box
        self.time_label = QLabel("Number of temporal frames:")

        self.time_input = QSpinBox()
        self.time_input.setRange(1, 1)
        self.time_input.setValue(1)
        self.time_input.setEnabled(False)
        self.time_input.valueChanged.connect(self.change_t)
        self.num_t = 1

        # Checkbox for 'Remove edges'
        self.remove_edges_checkbox = QCheckBox("Remove image edges")
        self.remove_edges_checkbox.setEnabled(False)
        self.remove_edges_checkbox.setToolTip(
            "Originally for Snouty deskewed images. If you see weird image edge artifacts, enable this.")

        # Run im button
        self.run_button = QPushButton(text="Run Nellie")
        self.run_button.clicked.connect(self.run_nellie)
        self.run_button.setEnabled(False)

        # Preprocess im button
        self.preprocess_button = QPushButton(text="Run preprocessing")
        self.preprocess_button.clicked.connect(self.run_preprocessing)
        self.preprocess_button.setEnabled(False)

        # Segment im button
        self.segment_button = QPushButton(text="Run segmentation")
        self.segment_button.clicked.connect(self.run_segmentation)
        self.segment_button.setEnabled(False)

        # Run mocap button
        self.mocap_button = QPushButton(text="Run mocap marking")
        self.mocap_button.clicked.connect(self.run_mocap)
        self.mocap_button.setEnabled(False)


        # Run tracking button
        self.track_button = QPushButton(text="Run tracking")
        self.track_button.clicked.connect(self.run_tracking)
        self.track_button.setEnabled(False)

        # Run reassign button
        self.reassign_button = QPushButton(text="Run voxel reassignment")
        self.reassign_button.clicked.connect(self.run_reassign)

        self.reassign_button.setEnabled(False)

        # Run feature extraction button
        self.feature_export_button = QPushButton(text="Run feature export")
        self.feature_export_button.clicked.connect(self.run_feature_export)
        self.feature_export_button.setEnabled(False)

        self.layout = QGridLayout()
        self.setLayout(self.layout)

        # Add buttons
        self.layout.addWidget(self.channel_label, 0, 0)
        self.layout.addWidget(self.channel_input, 0, 1)
        self.layout.addWidget(self.time_label, 1, 0)
        self.layout.addWidget(self.time_input, 1, 1)
        self.layout.addWidget(self.remove_edges_checkbox, 2, 1)

        self.layout.addWidget(QLabel("Run full pipeline"), 42, 0, 1, 2)
        self.layout.addWidget(self.run_button, 43, 0, 1, 2)

        self.layout.addWidget(QLabel("Run individual steps / Visualize"), 44, 0, 1, 2)
        self.layout.addWidget(self.preprocess_button, 45, 0, 1, 2)
        self.layout.addWidget(self.segment_button, 46, 0, 1, 2)
        self.layout.addWidget(self.mocap_button, 47, 0, 1, 2)
        self.layout.addWidget(self.track_button, 48, 0, 1, 2)
        self.layout.addWidget(self.feature_export_button, 49, 0, 1, 2)
        self.layout.addWidget(self.reassign_button, 50, 0, 1, 2)

        self.im_memmap = None
        self.raw_layer = None
        self.ray_layer = None
        self.ray_near_far = None
        self.point_layer = None

        self.animation_timer = None
        self.animation_start_time = None
        self.current_angles = None
        self.start_angle = None
        self.tracked_coords = None
        self.track_layer = None
        self.tracks = []

    def post_init(self):
        if not self.check_for_raw():
            return
        self.check_file_existence()
        self.remove_edges_checkbox.setEnabled(True)
        
    def check_file_existence(self):
        self.nellie.visualizer.check_file_existence()

        # set all other buttons to disabled first
        self.segment_button.setEnabled(False)
        self.mocap_button.setEnabled(False)
        self.track_button.setEnabled(False)
        self.reassign_button.setEnabled(False)
        self.feature_export_button.setEnabled(False)

        frangi_path = self.nellie.im_info.pipeline_paths['im_frangi']
        if os.path.exists(frangi_path):
            self.segment_button.setEnabled(True)
        else:
            self.segment_button.setEnabled(False)
            self.mocap_button.setEnabled(False)
            self.track_button.setEnabled(False)
            self.reassign_button.setEnabled(False)
            self.feature_export_button.setEnabled(False)
            return

        im_instance_label_path = self.nellie.im_info.pipeline_paths['im_instance_label']
        im_skel_relabelled_path = self.nellie.im_info.pipeline_paths['im_skel_relabelled']
        if os.path.exists(im_instance_label_path) and os.path.exists(im_skel_relabelled_path):
            self.mocap_button.setEnabled(True)
        else:
            self.mocap_button.setEnabled(False)
            self.track_button.setEnabled(False)
            self.reassign_button.setEnabled(False)
            self.feature_export_button.setEnabled(False)
            return

        im_marker_path = self.nellie.im_info.pipeline_paths['im_marker']
        if os.path.exists(im_marker_path):
            self.track_button.setEnabled(True)
        else:
            self.track_button.setEnabled(False)
            self.reassign_button.setEnabled(False)
            self.feature_export_button.setEnabled(False)
            return

        track_path = self.nellie.im_info.pipeline_paths['flow_vector_array']
        if os.path.exists(track_path):
            self.reassign_button.setEnabled(True)
            self.feature_export_button.setEnabled(True)
        else:
            self.reassign_button.setEnabled(False)
            self.feature_export_button.setEnabled(False)
            # return

        analysis_path = self.nellie.im_info.pipeline_paths['adjacency_maps']
        if os.path.exists(analysis_path):
            self.nellie.setTabEnabled(self.nellie.analysis_tab, True)
            self.nellie.analyzer.post_init()
        else:
            self.nellie.setTabEnabled(self.nellie.analysis_tab, False)


    def run_preprocessing(self):
        preprocessing = Filter(im_info=self.nellie.im_info, num_t=self.num_t,
                               remove_edges=self.remove_edges_checkbox.isChecked())
        preprocessing.run()

        self.check_file_existence()

    def run_segmentation(self):
        segmenting = Label(im_info=self.nellie.im_info, num_t=self.num_t)
        segmenting.run()
        networking = Network(im_info=self.nellie.im_info, num_t=self.num_t)
        networking.run()

        self.check_file_existence()

    def run_mocap(self):
        mocap_marking = Markers(im_info=self.nellie.im_info, num_t=self.num_t)
        mocap_marking.run()

        self.check_file_existence()

    def run_tracking(self):
        hu_tracking = HuMomentTracking(im_info=self.nellie.im_info, num_t=self.num_t)
        hu_tracking.run()

        self.check_file_existence()

    def run_reassign(self):
        vox_reassign = VoxelReassigner(im_info=self.nellie.im_info, num_t=self.num_t)
        vox_reassign.run()

        self.check_file_existence()

    def run_feature_export(self):
        hierarchy = Hierarchy(im_info=self.nellie.im_info, num_t=self.num_t)
        hierarchy.run()

        self.check_file_existence()

    def run_nellie(self):
        self.run_preprocessing()
        self.run_segmentation()
        self.run_mocap()
        self.run_tracking()
        self.run_reassign()
        self.run_feature_export()

        self.check_file_existence()

    def check_for_raw(self):
        self.preprocess_button.setEnabled(False)
        self.run_button.setEnabled(False)
        try:
            # todo implement lazy loading if wanted
            im_memmap = self.nellie.im_info.get_im_memmap(self.nellie.im_info.im_path)
            self.im_memmap = get_reshaped_image(im_memmap, self.num_t, self.nellie.im_info)
            self.preprocess_button.setEnabled(True)
            self.run_button.setEnabled(True)
            return True
        except Exception as e:
            logger.error(e)
            show_info(f"Could not open raw image: {e}")
            self.im_memmap = None
            return False

    def change_channel(self):
        if self.nellie.file_select.single:
            self.nellie.file_select.initialize_single_file()
        else:
            self.nellie.file_select.initialize_folder(self.nellie.file_select.filepath)

    def change_t(self):
        self.num_t = self.time_input.value()

    # def find_closest_voxel(self, near2, far2):
    #     near1, far1 = self.ray_near_far
    #     time_point = near1[0]
    #     near1 = near1[1:]
    #     far1 = far1[1:]
    #     near2 = near2[1:]
    #     far2 = far2[1:]
    #
    #     p1 = np.array(near1)
    #     d1 = np.array(far1) - p1
    #
    #     p2 = np.array(near2)
    #     d2 = np.array(far2) - p2
    #
    #     show_info(f"{p1}, {d1}, {p2}, {d2}")
    #
    #     # normalize
    #     d1 /= np.linalg.norm(d1)
    #     d2 /= np.linalg.norm(d2)
    #     show_info(f"{p1}, {d1}, {p2}, {d2}")
    #
    #     # cross
    #     cross_d1_d2 = np.cross(d1, d2)
    #
    #     # check non-parallel
    #     if np.linalg.norm(cross_d1_d2) < 1e-6:
    #         logger.error("Rays are parallel or coincident.")
    #         return
    #
    #     # Calculate the line segment connecting the two lines that is perpendicular to both
    #     # This is done by solving the system of linear equations formed by the two lines
    #     a = np.array([d1, -d2, cross_d1_d2]).T
    #     b = p2 - p1
    #     t, s, _ = np.linalg.solve(a, b)
    #
    #     # Calculate the closest points on the original lines
    #     closest_point_on_line1 = p1 + d1 * t
    #     closest_point_on_line2 = p2 + d2 * s
    #
    #     # Find the midpoint of the shortest line connecting the two rays
    #     midpoint = (closest_point_on_line1 + closest_point_on_line2) / 2
    #     # add time point to midpoint coords
    #     midpoint = np.insert(midpoint, 0, time_point)
    #     # Convert the midpoint to voxel coordinates (assumes the raw layer is the image)
    #     voxel_coords = np.round(self.raw_layer.world_to_data(midpoint)).astype(int)
    #     show_info(f"Midpoint: {midpoint}, Voxel coords: {voxel_coords}")
    #     # self.point_layer = self.viewer.add_points(voxel_coords, name='midpoint', face_color='red', size=2)
    #     self.tracked_coords = voxel_coords
    #
    #     # Check if the voxel_coords are within the bounds of the image
    #     if np.all(voxel_coords >= 0) and np.all(voxel_coords < self.im_memmap.shape):
    #         # Here you can do something with the voxel coordinates, like highlighting the voxel
    #         logger.info(f"Closest voxel at: {voxel_coords}")
    #     else:
    #         logger.error("Calculated voxel coordinates are out of bounds.")

    # def animate_camera_angle(self):
    #     # Calculate the time elapsed since the start of the animation
    #     time_elapsed = time.time() - self.animation_start_time
    #     duration = 0.5  # Total duration of the animation in seconds
    #
    #     if time_elapsed < duration:
    #         # Calculate the intermediate angle based on the elapsed time
    #         fraction = time_elapsed / duration
    #         delta_angle = 45 * fraction  # Change in angle should be 45 degrees
    #         new_angle = self.start_angle + delta_angle
    #
    #         # Update the camera angle
    #         self.viewer.camera.angles = (self.current_angles[0], self.current_angles[1], new_angle)
    #     else:
    #         # Animation is complete, set the final angle
    #         self.viewer.camera.angles = (self.current_angles[0], self.current_angles[1], self.start_angle + 45)
    #         self.animation_timer.stop()  # Stop the timer

    # def get_tracks(self):
    #     if self.tracked_coords is None:
    #         return
    #     t_current = self.tracked_coords[0]
    #     coord_to_track = np.asarray([self.tracked_coords[1:]])
    #     tracks_forward, track_props_forward = interpolate_all_forward(coord_to_track, t_current,
    #                                                                   self.num_t - 1, self.nellie.im_info)
    #     tracks_back, track_props_back = interpolate_all_backward(coord_to_track, t_current,
    #                                                             1, self.nellie.im_info)
    #     self.track_layer = self.viewer.add_tracks(tracks_forward, properties=track_props_forward, name='tracks', scale=self.scale)
    #     self.track_layer = self.viewer.add_tracks(tracks_back, properties=track_props_back, name='tracks', scale=self.scale)

    # def on_click(self, layer, event):  # https://napari.org/dev/guides/3D_interactivity.html
    #     near, far = None, None
    #     if event.button == 1 and event.is_dragging is False and 'Alt' in event.modifiers:
    #         near, far = layer.get_ray_intersections(event.position, event.view_direction, event.dims_displayed)
    #     if near is None or far is None:
    #         return
    #
    #     # add line to viewer
    #     if self.ray_layer is None:
    #         self.ray_near_far = (near, far)
    #         self.ray_layer = self.viewer.add_shapes(
    #             np.asarray([[*near], [*far]]),
    #             shape_type='line',
    #             edge_color='red',
    #             edge_width=1,
    #             name='ray',
    #             opacity=0.1, scale=self.scale
    #         )
    #         self.viewer.layers.selection.active = self.viewer.layers[0]
    #
    #         # Store the current angles and start angle for the animation
    #         self.current_angles = self.viewer.camera.angles
    #         self.start_angle = self.current_angles[2]
    #
    #         # Start the animation
    #         self.animation_start_time = time.time()
    #         self.animation_timer = QTimer()
    #         self.animation_timer.timeout.connect(self.animate_camera_angle)
    #         self.animation_timer.start(1000 // 60)  # Update at 60 FPS
    #     else:
    #         self.find_closest_voxel(near, far)
    #         # remove ray layer
    #         self.viewer.layers.remove(self.ray_layer)
    #         self.ray_layer = None
    #         self.ray_near_far = None
    #         self.get_tracks()


if __name__ == "__main__":
    import napari
    viewer = napari.Viewer()
    napari.run()



import heapq

import numpy as np
from scipy.spatial import cKDTree

from nellie import logger
from nellie.im_info.im_info import ImInfo
from nellie.tracking.flow_interpolation import FlowInterpolator
from nellie.utils.general import get_reshaped_image


class VoxelReassigner:
    def __init__(self, im_info: ImInfo, num_t=None):
        self.im_info = im_info
        self.num_t = num_t
        if num_t is None and not self.im_info.no_t:
            self.num_t = im_info.shape[im_info.axes.index('T')]
        self.flow_interpolator_fw = FlowInterpolator(im_info)
        self.flow_interpolator_bw = FlowInterpolator(im_info, forward=False)

        self.running_matches = []

        self.voxel_matches_path = None
        self.branch_label_memmap = None
        self.obj_label_memmap = None
        self.reassigned_branch_memmap = None
        self.reassigned_obj_memmap = None

        self.debug = None

    def _match_forward(self, flow_interpolator, vox_prev, vox_next, t):
        vectors_interpx_prev = flow_interpolator.interpolate_coord(vox_prev, t)
        if vectors_interpx_prev is None:
            return [], [], []
        # only keep voxels that are not nan
        kept_prev_vox_idxs = ~np.isnan(vectors_interpx_prev).any(axis=1)
        # only keep vectors where the voxel is not nan
        vectors_interpx_prev = vectors_interpx_prev[kept_prev_vox_idxs]
        # get centroids in t1 from voxels in t0 + interpolated flow at that voxel
        vox_prev_kept = vox_prev[kept_prev_vox_idxs]
        centroids_next_interpx = vox_prev_kept + vectors_interpx_prev
        if len(centroids_next_interpx) == 0:
            return [], [], []
        # now we have estimated centroids in t1 (centroids_next_interpx) and linked voxels in t0 (vox_prev[kept_prev_vox_idxs]).
        # we then have to match t1 voxels (vox_next) to estimated t1 centroids (centroids_next_interpx)
        match_dist, matched_idx = self._match_voxels_to_centroids(vox_next, centroids_next_interpx)
        vox_matched_to_centroids = vox_next[matched_idx.tolist()]
        # then link those t1 voxels back to the t0 voxels
        # now we have linked t0 voxels (vox_prev_kept) to t1 voxels (vox_matched_to_centroids)
        # but we have to make sure the link is within a distance constraint.
        vox_prev_matched_valid, vox_next_matched_valid, distances_valid = self._distance_threshold(
            vox_prev_kept, vox_matched_to_centroids
        )
        return vox_prev_matched_valid, vox_next_matched_valid, distances_valid

    def _match_backward(self, flow_interpolator, vox_next, vox_prev, t):
        # interpolate flow vectors to all voxels in t1 from centroids derived from t0 centroids + t0 flow vectors
        vectors_interpx_prev = flow_interpolator.interpolate_coord(vox_next, t)
        if vectors_interpx_prev is None:
            return [], [], []
        # only keep voxels that are not nan
        kept_next_vox_idxs = ~np.isnan(vectors_interpx_prev).any(axis=1)
        # only keep vectors where the voxel is not nan
        vectors_interpx_prev = vectors_interpx_prev[kept_next_vox_idxs]
        # get centroids in t0 from voxels in t1 - interpolated flow (from t0 to t1) at that voxel
        vox_next_kept = vox_next[kept_next_vox_idxs]
        centroids_prev_interpx = vox_next_kept - vectors_interpx_prev
        if len(centroids_prev_interpx) == 0:
            return [], [], []
        # now we have estimated centroids in t0 (centroids_prev_interpx) and linked voxels in t1 (vox_next[kept_next_vox_idxs]).
        # we then have to match t0 voxels (vox_prev) to estimated t0 centroids (centroids_prev_interpx)
        match_dist, matched_idx = self._match_voxels_to_centroids(vox_prev, centroids_prev_interpx)
        vox_matched_to_centroids = vox_prev[matched_idx.tolist()]
        # then link those t1 voxels (vox_next_kept) back to the t0 voxels (vox_matched_to_centroids).
        # but we have to make sure the link is within a distance constraint.
        vox_prev_matched_valid, vox_next_matched_valid, distances_valid = self._distance_threshold(
            vox_matched_to_centroids, vox_next_kept
        )
        return vox_prev_matched_valid, vox_next_matched_valid, distances_valid

    def _match_voxels_to_centroids(self, coords_real, coords_interpx):
        coords_interpx = np.array(coords_interpx) * self.flow_interpolator_fw.scaling
        coords_real = np.array(coords_real) * self.flow_interpolator_fw.scaling
        tree = cKDTree(coords_real)
        dist, idx = tree.query(coords_interpx, k=1, workers=-1)
        return dist, idx

    def _assign_unique_matches(self, vox_prev_matches, vox_next_matches, distances):
        # create a dict where the key is a voxel in t1, and the value is a list of distances and t0 voxels matched to it
        vox_next_dict = {}
        for match_idx, match_next in enumerate(vox_next_matches):
            match_next_tuple = tuple(match_next)
            if match_next_tuple not in vox_next_dict.keys():
                vox_next_dict[match_next_tuple] = [[], []]
            vox_next_dict[match_next_tuple][0].append(distances[match_idx])
            vox_next_dict[match_next_tuple][1].append(vox_prev_matches[match_idx])

        # now assign matches based on the t1 voxel's closest (in distance) matched t0 voxel
        vox_prev_matches_final = []
        vox_next_matches_final = []
        for match_next_tuple, (distance_match_list, vox_prev_match_list) in vox_next_dict.items():
            if len(distance_match_list) == 1:
                vox_prev_matches_final.append(vox_prev_match_list[0])
                vox_next_matches_final.append(match_next_tuple)
                continue
            min_idx = np.argmin(distance_match_list)
            vox_prev_matches_final.append(vox_prev_match_list[min_idx])
            vox_next_matches_final.append(match_next_tuple)

        # create a priority queue with (distance, prev_voxel, next_voxel) tuples
        priority_queue = [(distances[i], tuple(vox_prev_matches[i]), tuple(vox_next_matches[i]))
                          for i in range(len(distances))]
        heapq.heapify(priority_queue)  # Convert list to a heap in-place

        assigned_prev = set()
        assigned_next = set()
        vox_prev_matches_final = []
        vox_next_matches_final = []

        while priority_queue:
            # pop the smallest distance tuple from the heap
            distance, prev_voxel, next_voxel = heapq.heappop(priority_queue)

            if prev_voxel not in assigned_prev or next_voxel not in assigned_next:
                # if neither of the voxels has been assigned, then assign them
                vox_prev_matches_final.append(prev_voxel)
                vox_next_matches_final.append(next_voxel)
                assigned_prev.add(prev_voxel)
                assigned_next.add(next_voxel)

        return vox_prev_matches_final, vox_next_matches_final

    def _distance_threshold(self, vox_prev_matched, vox_next_matched):
        distances = np.linalg.norm((vox_prev_matched - vox_next_matched) * self.flow_interpolator_fw.scaling, axis=1)
        distance_mask = distances < self.flow_interpolator_fw.max_distance_um
        vox_prev_matched_valid = vox_prev_matched[distance_mask]
        vox_next_matched_valid = vox_next_matched[distance_mask]
        distances_valid = distances[distance_mask]
        return vox_prev_matched_valid, vox_next_matched_valid, distances_valid

    def match_voxels(self, vox_prev, vox_next, t):
        # forward interpolation:
        # from t0 voxels and interpolated flow, get t1 centroids.
        #  match nearby t1 voxels to t1 centroids, which are linked to t0 voxels.
        logger.debug(f'Forward voxel matching for t: {t}')
        vox_prev_matches_fw, vox_next_matches_fw, distances_fw = self._match_forward(
            self.flow_interpolator_fw, vox_prev, vox_next, t
        )

        # backward interpolation:
        # from t0 centroids and real flow, get t1 centroids.
        #  interpolate flow at nearby t1 voxels. subtract flow from voxels to get t0 centroids.
        #  match nearby t0 voxels to t0 centroids, which are linked to t1 voxels.
        logger.debug(f'Backward voxel matching for t: {t}')
        vox_prev_matches_bw, vox_next_matches_bw, distances_bw = self._match_backward(
            self.flow_interpolator_bw, vox_next, vox_prev, t + 1
        )

        logger.debug(f'Assigning unique matches for t: {t}')
        vox_prev_matches = np.concatenate([vox_prev_matches_fw, vox_prev_matches_bw])
        vox_next_matches = np.concatenate([vox_next_matches_fw, vox_next_matches_bw])
        distances = np.concatenate([distances_fw, distances_bw])

        vox_prev_matches_unique, vox_next_matches_unique = self._assign_unique_matches(vox_prev_matches,
                                                                                       vox_next_matches, distances)

        vox_next_matches_unique = np.array(vox_next_matches_unique)
        vox_next_matched_tuples = set([tuple(coord) for coord in vox_next_matches_unique])
        vox_next_unmatched = np.array([coord for coord in vox_next if tuple(coord) not in vox_next_matched_tuples])
        unmatched_diff = np.inf
        while unmatched_diff:
            num_unmatched = len(vox_next_unmatched)
            if num_unmatched == 0:
                break
            tree = cKDTree(vox_next_matches_unique * self.flow_interpolator_fw.scaling)
            dists, idxs = tree.query(vox_next_unmatched * self.flow_interpolator_fw.scaling, k=1, workers=-1)
            unmatched_matches = np.array([
                [vox_prev_matches_unique[idx], vox_next_unmatched[i]]
                for i, idx in enumerate(idxs) if dists[i] < self.flow_interpolator_fw.max_distance_um
            ])
            if len(unmatched_matches) == 0:
                break
            # add unmatched matches to coords_matched
            vox_prev_matches_unique = np.concatenate([vox_prev_matches_unique, unmatched_matches[:, 0]])
            vox_next_matches_unique = np.concatenate([vox_next_matches_unique, unmatched_matches[:, 1]])
            vox_next_matched_tuples = set([tuple(coord) for coord in vox_next_matches_unique])
            vox_next_unmatched = np.array([coord for coord in vox_next if tuple(coord) not in vox_next_matched_tuples])
            new_num_unmatched = len(vox_next_unmatched)
            unmatched_diff = num_unmatched - new_num_unmatched
            logger.debug(f'Reassigned {unmatched_diff}/{num_unmatched} unassigned voxels. '
                         f'{new_num_unmatched} remain.')
        return np.array(vox_prev_matches_unique), np.array(vox_next_matches_unique)

    def _get_t(self):
        if self.num_t is None:
            if self.im_info.no_t:
                self.num_t = 1
            else:
                self.num_t = self.im_info.shape[self.im_info.axes.index('T')]
        else:
            return

    def _allocate_memory(self):
        logger.debug('Allocating memory for voxel reassignment.')
        self.voxel_matches_path = self.im_info.pipeline_paths['voxel_matches']

        branch_label_memmap = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_skel_relabelled'])
        obj_label_memmap = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_instance_label'])
        self.branch_label_memmap = get_reshaped_image(branch_label_memmap, self.num_t, self.im_info)
        self.obj_label_memmap = get_reshaped_image(obj_label_memmap, self.num_t, self.im_info)
        self.shape = self.branch_label_memmap.shape

        reassigned_branch_label_path = self.im_info.pipeline_paths['im_branch_label_reassigned']
        self.reassigned_branch_memmap = self.im_info.allocate_memory(reassigned_branch_label_path, shape=self.shape,
                                                                     dtype='int32',
                                                                     description='branch label reassigned',
                                                                     return_memmap=True)

        reassigned_obj_label_path = self.im_info.pipeline_paths['im_obj_label_reassigned']
        self.reassigned_obj_memmap = self.im_info.allocate_memory(reassigned_obj_label_path, shape=self.shape,
                                                                  dtype='int32',
                                                                  description='object label reassigned',
                                                                  return_memmap=True)

    def _run_frame(self, t, all_mask_coords, reassigned_memmap):
        logger.info(f'Reassigning pixels in frame {t + 1} of {self.num_t - 1}')

        vox_prev = all_mask_coords[t]
        vox_next = all_mask_coords[t + 1]
        if len(vox_prev) == 0 or len(vox_next) == 0:
            return True

        matched_prev, matched_next = self.match_voxels(vox_prev, vox_next, t)
        if len(matched_prev) == 0:
            return True
        matched_prev = matched_prev.astype('uint16')
        matched_next = matched_next.astype('uint16')

        self.running_matches.append([matched_prev, matched_next])

        reassigned_memmap[t + 1][tuple(matched_next.T)] = reassigned_memmap[t][tuple(matched_prev.T)]

        return False

    def _run_reassignment(self, label_type):
        # todo, be able to specify which frame to start at.
        if label_type == 'branch':
            label_memmap = self.branch_label_memmap
            reassigned_memmap = self.reassigned_branch_memmap
        elif label_type == 'obj':
            label_memmap = self.obj_label_memmap
            reassigned_memmap = self.reassigned_obj_memmap
        else:
            raise ValueError('label_type must be "branch" or "obj".')
        vox_prev = np.argwhere(label_memmap[0] > 0)
        reassigned_memmap[0][tuple(vox_prev.T)] = label_memmap[0][tuple(vox_prev.T)]
        all_mask_coords = [np.argwhere(label_memmap[t] > 0) for t in range(self.num_t)]

        for t in range(self.num_t - 1):
            no_matches = self._run_frame(t, all_mask_coords, reassigned_memmap)

            if no_matches:
                break

    def run(self):
        if self.im_info.no_t:
            return
        self._get_t()
        self._allocate_memory()
        self._run_reassignment('branch')
        self._run_reassignment('obj')
        # save running matches to npy
        np.save(self.voxel_matches_path, np.array(self.running_matches, dtype=object))


if __name__ == "__main__":
    im_path = r"D:\test_files\nelly_smorgasbord\deskewed-iono_pre.ome.tif"
    im_info = ImInfo(im_path)
    num_t = 3
    run_obj = VoxelReassigner(im_info, num_t=num_t)
    run_obj.run()

    import pickle

    # This section seems random, but it allows for finding links between any level of the hierarchy to any
    #  other level in the hierarchy at any time point via lots of dot products.
    edges_loaded = pickle.load(open(im_info.pipeline_paths['adjacency_maps'], "rb"))

    mask_01 = run_obj.obj_label_memmap[:2] > 0
    mask_voxels_0 = np.argwhere(mask_01[0])
    mask_voxels_1 = np.argwhere(mask_01[1])

    t0_coords_in_mask_0 = {tuple(coord): idx for idx, coord in enumerate(mask_voxels_0)}
    t1_coords_in_mask_1 = {tuple(coord): idx for idx, coord in enumerate(mask_voxels_1)}

    idx_matches_0 = [t0_coords_in_mask_0[tuple(coord)] for coord in run_obj.running_matches[0][0]]
    idx_matches_1 = [t1_coords_in_mask_1[tuple(coord)] for coord in run_obj.running_matches[0][1]]
    # sort based on idx_matches_0
    sorted_idx_matches_0_order = np.argsort(idx_matches_0)
    sorted_idx_matches_0 = np.array(idx_matches_0)[sorted_idx_matches_0_order]
    sorted_idx_matches_1 = np.array(idx_matches_1)[sorted_idx_matches_0_order]

    v_t = np.zeros((len(mask_voxels_0), len(mask_voxels_1)), dtype=np.uint16)
    v_t[sorted_idx_matches_0, sorted_idx_matches_1] = True

    b_v = edges_loaded['b_v'][0].astype(np.uint16)
    # dot product b_v and v_t
    import cupy as cp


    def dot_product_in_chunks(a, b, chunk_size=100):
        result = cp.zeros((a.shape[0], b.shape[1]), dtype=cp.uint8)
        for start_row in range(0, a.shape[1], chunk_size):
            print(start_row, start_row + chunk_size)
            end_row = start_row + chunk_size
            v_t_chunk = b[start_row:end_row, :]
            b_v_chunk = a[:, start_row:end_row]
            result += cp.dot(b_v_chunk, v_t_chunk)  # Adjust this line as per your logic
        return result


    # Convert your numpy arrays to cupy arrays
    v_t_cp = cp.array(v_t, dtype=cp.uint8)
    b_v_cp = cp.array(b_v, dtype=cp.uint8)

    # Perform dot product in chunks
    b0_v1_cp = dot_product_in_chunks(b_v_cp, v_t_cp)

    # Convert the result back to a numpy array if needed
    b1_v1 = edges_loaded['b_v'][1].astype(np.uint16)
    b1_v1_cp = cp.array(b1_v1, dtype=cp.uint8)
    b0_b1_cp = dot_product_in_chunks(b0_v1_cp, b1_v1_cp.T)
    b0_b1 = cp.asnumpy(b0_b1_cp)
    # b0_b1 are the new edges between branches in time 0 and time 1, values are the number of voxels in common between (aka weighting to give the edges)

    # find the indices of the maximum value in each col
    max_idx = np.argmax(b0_b1, axis=0) + 1

    mask_branches = np.zeros(mask_01.shape, dtype=np.uint16)
    branch_labels_0 = np.argmax(b_v.T, axis=1)
    branch_labels_1 = np.argmax(b1_v1.T, axis=1)

    mask_branches[0][tuple(mask_voxels_0.T)] = branch_labels_0 + 1

    # replace any non-zero values in b1_v1 with the max_idx
    new_branch_labels_1 = max_idx[branch_labels_1]
    mask_branches[1][tuple(mask_voxels_1.T)] = new_branch_labels_1 + 1
    # these branches are relabelled by t0 branch labels.

    # lets do this with nodes, too
    n_v = edges_loaded['n_v'][0].astype(np.uint16)
    n_v_cp = cp.array(n_v, dtype=cp.uint8)
    n0_v1_cp = dot_product_in_chunks(n_v_cp, v_t_cp)
    n1_v1 = edges_loaded['n_v'][1].astype(np.uint16)
    n1_v1_cp = cp.array(n1_v1, dtype=cp.uint8)
    n0_n1_cp = dot_product_in_chunks(n0_v1_cp, n1_v1_cp.T)

    n0_n1 = cp.asnumpy(n0_n1_cp)
    max_idx_n = np.argmax(n0_n1, axis=0) + 1

    mask_nodes = np.zeros(mask_01.shape, dtype=np.uint16)
    node_labels_0 = np.argmax(n_v.T, axis=1)
    node_labels_1 = np.argmax(n1_v1.T, axis=1)
    mask_nodes[0][tuple(mask_voxels_0.T)] = node_labels_0 + 1
    new_node_labels_1 = max_idx_n[node_labels_1]
    mask_nodes[1][tuple(mask_voxels_1.T)] = new_node_labels_1 + 1

    # # todo useful for getting continuous tracks for voxels
    # matches_t0_t1 = run_obj.running_matches[0][1]
    # matches_t1_t2 = run_obj.running_matches[1][0]
    #
    # t1_coords_in_t0_t1 = {tuple(coord): idx for idx, coord in enumerate(matches_t0_t1)}
    # t1_coords_in_t1_t2 = {tuple(coord): idx for idx, coord in enumerate(matches_t1_t2)}
    #
    # t0_coords_in_t0_t1 = {tuple(coord): idx for idx, coord in enumerate(run_obj.running_matches[0][0])}
    #
    # # Create the continuous track list
    # continuous_tracks = []
    # for t1_coord, t0_idx in t1_coords_in_t0_t1.items():
    #     if t1_coord in t1_coords_in_t1_t2:
    #         t0_coord = run_obj.running_matches[0][0][t0_idx]
    #         t2_idx = t1_coords_in_t1_t2[t1_coord]
    #         t2_coord = run_obj.running_matches[1][1][t2_idx]
    #         continuous_tracks.append([t0_coord, t1_coord, t2_coord])
    #
    # napari_tracks = []
    # for i, track in enumerate(continuous_tracks):
    #     napari_tracks.append([i, 0, track[0][0], track[0][1], track[0][2]])
    #     napari_tracks.append([i, 1, track[1][0], track[1][1], track[1][2]])
    #     napari_tracks.append([i, 2, track[2][0], track[2][1], track[2][2]])



from nellie import xp, device_type


def otsu_effectiveness(image, inter_variance):
    # flatten image and create histogram
    flattened_image = image.flatten()
    sigma_total_squared = xp.var(flattened_image)
    normalized_sigma_B_squared = inter_variance / sigma_total_squared
    return normalized_sigma_B_squared


def otsu_threshold(matrix, nbins=256):
    # gpu version of skimage.filters.threshold_otsu
    counts, bin_edges = xp.histogram(matrix.reshape(-1), bins=nbins, range=(matrix.min(), matrix.max()))
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2.
    counts = counts / xp.sum(counts)

    weight1 = xp.cumsum(counts)
    mean1 = xp.cumsum(counts * bin_centers) / weight1
    if device_type == 'mps':
        weight2 = xp.cumsum(xp.flip(counts, dims=[0]))
        weight2 = xp.flip(weight2, dims=[0])
        flipped_counts_bin_centers = xp.flip(counts * bin_centers, dims=[0])
        cumsum_flipped = xp.cumsum(flipped_counts_bin_centers)
        mean2 = xp.flip(cumsum_flipped / xp.flip(weight2, dims=[0]), dims=[0])
    else:
        weight2 = xp.cumsum(counts[::-1])[::-1]
        mean2 = (xp.cumsum((counts * bin_centers)[::-1]) / weight2[::-1])[::-1]

    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:]) ** 2

    idx = xp.argmax(variance12)
    threshold = bin_centers[idx]

    return threshold, variance12[idx]


def triangle_threshold(matrix, nbins=256):
    # gpu version of skimage.filters.threshold_triangle
    hist, bin_edges = xp.histogram(matrix.reshape(-1), bins=nbins, range=(xp.min(matrix), xp.max(matrix)))
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2.
    hist = hist / xp.sum(hist)

    arg_peak_height = xp.argmax(hist)
    peak_height = hist[arg_peak_height]
    arg_low_level, arg_high_level = xp.flatnonzero(hist)[[0, -1]]

    flip = arg_peak_height - arg_low_level < arg_high_level - arg_peak_height
    if flip:
        if device_type == 'mps':
            hist = xp.flip(hist, dims=[0])
        else:
            hist = xp.flip(hist, axis=0)

            # todo check this
        arg_low_level = nbins - arg_high_level - 1
        arg_peak_height = nbins - arg_peak_height - 1
    del (arg_high_level)

    width = arg_peak_height - arg_low_level
    x1 = xp.arange(width)
    y1 = hist[x1 + arg_low_level]

    norm = xp.sqrt(peak_height ** 2 + width ** 2)
    peak_height = peak_height / norm
    width = width / norm

    length = peak_height * x1 - width * y1
    arg_level = xp.argmax(length) + arg_low_level

    if flip:
        arg_level = nbins - arg_level - 1

    return bin_centers[arg_level]



import datetime
import os

import napari
from qtpy.QtWidgets import QWidget, QGridLayout, QLabel, QPushButton, QSpinBox, QCheckBox, QLineEdit, QMessageBox, \
    QFileDialog
from napari.utils.notifications import show_info
from tifffile import tifffile

from nellie.im_info.im_info import ImInfo
from nellie.utils.general import get_reshaped_image


class NellieFileSelect(QWidget):
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', nellie, parent=None):
        super().__init__(parent)
        self.nellie = nellie
        self.filepath = None
        self.im_info = None
        self.im_info_valid = False

        self.num_frames_max = None
        self.num_channels = 1

        self.viewer = napari_viewer
        self.viewer.title = 'Nellie Napari'
        self.setLayout(QGridLayout())

        # Add text showing what filepath is selected
        self.filepath_text = QLabel(text="No file selected.")
        self.filepath_text.setWordWrap(True)

        # Filepath selector button
        self.filepath_button = QPushButton(text="Select File")
        self.filepath_button.clicked.connect(self.select_filepath)
        self.filepath_button.setEnabled(True)
        # Folder selector button
        self.folder_button = QPushButton(text="Select Folder")
        self.folder_button.clicked.connect(self.select_folder)
        self.folder_button.setEnabled(True)

        # reset button
        self.reset_button = QPushButton(text="Reset")
        self.reset_button.clicked.connect(self.reset)
        self.reset_button.setEnabled(True)

        # dims buttons
        self.label_t = QLabel("T resolution (s):")
        self.dim_t_button = QLineEdit(self)
        self.dim_t_button.textChanged.connect(self.handle_t_changed)
        self.dim_t = None
        self.label_z = QLabel("Z resolution (um):")
        self.dim_z_button = QLineEdit(self)
        self.dim_z_button.textChanged.connect(self.handle_z_changed)
        self.dim_z = None
        self.label_xy = QLabel("X & Y resolution (um):")
        self.dim_xy_button = QLineEdit(self)
        self.dim_xy_button.textChanged.connect(self.handle_xy_changed)
        self.dim_xy = None

        self.dim_sizes = {'T': None, 'Z': None, 'X': None, 'Y': None}
        self.dim_sizes_changed = False

        self.dim_order_button = QLineEdit(self)
        self.dim_order_button.setText("Select file.")
        self.dim_order_button.setToolTip("Accepted axes: ['TZYX', 'TYX', 'TZCYX', 'TCYX', 'TCZYX', 'ZYX', 'YX', 'CYX', 'CZYX', 'ZCYX']")
        self.dim_order = None
        self.dim_order_button.setEnabled(False)
        # self.dim_order_button.textChanged.connect(self.handleDimOrderChanged)

        self.set_dims_button = QPushButton(text="Set dimension order")
        self.set_dims_button.clicked.connect(self.handle_dim_order_changed)
        self.set_dims_button.setEnabled(False)

        self.set_res_button = QPushButton(text="Set new resolutions")
        self.set_res_button.clicked.connect(self.handle_res_changed)
        self.set_res_button.setEnabled(False)

        self.open_preview_button = QPushButton(text="Open preview")
        self.open_preview_button.setToolTip("Open two timepoints to check for valid dimension order and resolutions.")
        self.open_preview_button.clicked.connect(self.open_preview)
        self.open_preview_button.setEnabled(False)


        self.layout().addWidget(self.filepath_button, 1, 0, 1, 1)
        self.layout().addWidget(self.folder_button, 1, 1, 1, 1)
        self.layout().addWidget(self.filepath_text, 2, 0, 1, 2)
        self.layout().addWidget(QLabel("Dimension order: "), 3, 0, 1, 1)
        self.layout().addWidget(self.dim_order_button, 3, 1, 1, 1)
        self.layout().addWidget(self.set_dims_button, 4, 1, 1, 1)
        self.layout().addWidget(self.label_t, 5, 0, 1, 1)
        self.layout().addWidget(self.dim_t_button, 5, 1, 1, 1)
        self.layout().addWidget(self.label_z, 6, 0, 1, 1)
        self.layout().addWidget(self.dim_z_button, 6, 1, 1, 1)
        self.layout().addWidget(self.label_xy, 7, 0, 1, 1)
        self.layout().addWidget(self.dim_xy_button, 7, 1, 1, 1)
        self.layout().addWidget(self.set_res_button, 8, 1, 1, 1)
        self.layout().addWidget(self.open_preview_button, 49, 0, 1, 2)
        self.layout().addWidget(self.reset_button, 50, 0, 2, 2)

        self.nellie_viewer = None
        self.batch_mode = None
        self.nellie_analyzer = None

        self.single = None

    def open_preview(self):
        im_memmap = self.im_info.get_im_memmap(self.im_info.im_path)
        im_memmap_loaded = get_reshaped_image(im_memmap, 2, self.im_info)
        if self.im_info.no_z:
            scale = (self.im_info.dim_sizes['Y'], self.im_info.dim_sizes['X'])
        else:
            scale = (self.im_info.dim_sizes['Z'], self.im_info.dim_sizes['Y'], self.im_info.dim_sizes['X'])
        self.viewer.add_image(im_memmap_loaded, name=self.im_info.basename_no_ext,
                              scale=scale, blending='additive')
        self.viewer.scale_bar.visible = True
        self.viewer.scale_bar.unit = 'um'
        # set dims to 3d
        if self.im_info.no_z:
            self.viewer.dims.ndisplay = 2
        else:
            self.viewer.dims.ndisplay = 3

    def handle_res_changed(self, text):
        if self.dim_t is not None:
            self.dim_sizes['T'] = self.dim_t
        if self.dim_z is not None:
            self.dim_sizes['Z'] = self.dim_z
        if self.dim_xy is not None:
            self.dim_sizes['X'] = self.dim_xy
            self.dim_sizes['Y'] = self.dim_xy
        self.dim_sizes_changed = True
        if self.single:
            self.initialize_single_file()
        else:
            self.initialize_folder(self.filepath)

    def handle_dim_order_changed(self, text):
        self.dim_order = self.dim_order_button.text()
        if self.single:
            self.initialize_single_file()
        else:
            self.initialize_folder(self.filepath)

    def handle_t_changed(self, text):
        self.dim_t = self.handle_float(text)

    def handle_z_changed(self, text):
        self.dim_z = self.handle_float(text)

    def handle_xy_changed(self, text):
        self.dim_xy = self.handle_float(text)

    def handle_float(self, text):
        try:
            # Convert the text to float, store to respective dim
            if text is None:
                return None
            value = float(text)
            return value
        except ValueError:
            # Handle the case where conversion to float fails
            show_info("Please enter a valid number")
            return None

    def check_dims(self):
        self.set_dims_button.setEnabled(True)
        self.dim_order_button.setEnabled(True)
        self.dim_order_button.setText(self.im_info.axes)
        accepted_axes = ['TZYX', 'TYX', 'TZCYX', 'TCYX', 'TCZYX', 'ZYX', 'YX', 'CYX', 'CZYX', 'ZCYX']
        if self.dim_order not in accepted_axes:
            self.im_info.axes_valid = False
        else:
            self.im_info.axes_valid = True

        if not self.im_info.axes_valid:
            show_info("Invalid dimension order. Please check the dimension order.")
            self.dim_order_button.setStyleSheet("background-color: red")
            return
        self.dim_order_button.setStyleSheet("background-color: green")

    def check_axes(self):
        # by default enable all the res buttons
        self.set_res_button.setEnabled(True)
        self.dim_t_button.setEnabled(True)
        self.dim_z_button.setEnabled(True)
        self.dim_xy_button.setEnabled(True)

        self.dims_valid = True

        if self.im_info.no_t:
            self.dim_t_button.setEnabled(False)
        elif self.im_info.dim_sizes['T'] is not None:
            self.dim_t_button.setText(str(self.im_info.dim_sizes['T']))
            self.dim_t = self.im_info.dim_sizes['T']
            self.dim_t_button.setStyleSheet("background-color: green")
        else:
            self.dim_t_button.setStyleSheet("background-color: red")
            self.dims_valid = False

        if self.im_info.no_z:
            self.dim_z_button.setEnabled(False)
        elif self.im_info.dim_sizes['Z'] is not None:
            self.dim_z_button.setText(str(self.im_info.dim_sizes['Z']))
            self.dim_z = self.im_info.dim_sizes['Z']
            self.dim_z_button.setStyleSheet("background-color: green")
        else:
            self.dim_z_button.setStyleSheet("background-color: red")
            self.dims_valid = False

        if self.im_info.dim_sizes['X'] is not None:
            self.dim_xy_button.setText(str(self.im_info.dim_sizes['X']))
            self.dim_xy = self.im_info.dim_sizes['X']
            self.dim_xy_button.setStyleSheet("background-color: green")
        elif self.im_info.dim_sizes['Y'] is not None:
            self.dim_xy_button.setText(str(self.im_info.dim_sizes['Y']))
            self.dim_xy = self.im_info.dim_sizes['Y']
            self.dim_xy_button.setStyleSheet("background-color: green")
        else:
            self.dim_xy_button.setStyleSheet("background-color: red")
            self.dims_valid = False

    def check_manual_params(self):
        if self.dim_sizes_changed:
            dim_sizes = self.dim_sizes
        else:
            dim_sizes = None

        if self.dim_order is None:
            dim_order = ''
        else:
            dim_order = self.dim_order
        return dim_sizes, dim_order

    def check_im_info_valid(self):
        self.open_preview_button.setEnabled(False)
        self.check_dims()
        self.check_axes()
        if not self.dims_valid:
            show_info("Axes resolutions not set.")
            return False
        if not self.im_info.axes_valid:
            show_info("Dimension order is invalid.")
            return False
        self.open_preview_button.setEnabled(True)
        return True

    def set_max_frames(self):
        if self.im_info.no_t:
            self.num_frames_max = 1
        else:  # the index of self.im_info.shape that corresponds to the index of 'T' in self.im_info.axes
            self.num_frames_max = self.im_info.shape[self.im_info.axes.index('T')]

    def set_num_channels(self):
        if self.im_info.no_c:
            self.num_channels = 1
        else:  # the index of self.im_info.shape that corresponds to the index of 'C' in self.im_info.axes
            self.num_channels = self.im_info.shape[self.im_info.axes.index('C')]

    def initialize_single_file(self):
        dim_sizes, dim_order = self.check_manual_params()
        # open the file and load its info
        self.im_info = ImInfo(self.filepath, ch=self.nellie.processor.channel_input.value(), dimension_order=dim_order, dim_sizes=dim_sizes)
        self.nellie.im_info = self.im_info
        self.dim_order = self.im_info.axes

        # check if the dimension order and axes resolutions are valid
        self.im_info_valid = self.check_im_info_valid()

        self.post_file_selection()

    def initialize_folder(self, filepath):
        dim_sizes, dim_order = self.check_manual_params()
        # open the folder and load all the files it can
        filenames = os.listdir(filepath)
        self.nellie.valid_files = []
        for filename in filenames:
            try:
                im_info = ImInfo(os.path.join(filepath, filename), ch=self.nellie.processor.channel_input.value(),
                                 dim_sizes=dim_sizes, dimension_order=dim_order)
                self.nellie.valid_files.append(im_info)
            except:
                pass
        self.nellie.im_info = self.im_info

        self.im_info_valid = True
        for file in self.nellie.valid_files:
            self.im_info = file
            if not self.check_im_info_valid():
                self.im_info_valid = False
                break

        self.im_info = self.nellie.valid_files[0]
        self.post_file_selection()

    def check_analysis_valid(self):
        self.nellie.setTabEnabled(self.nellie.analysis_tab, False)
        if self.single and os.path.exists(self.im_info.pipeline_paths['adjacency_maps']):
            self.nellie.setTabEnabled(self.nellie.analysis_tab, True)
            self.nellie.analyzer.post_init()

    def disable_file_selection(self):
        self.filepath_button.setEnabled(False)
        self.folder_button.setEnabled(False)

    def enable_time_options(self):
        self.nellie.processor.time_input.setEnabled(False)
        self.nellie.batch_mode.time_input.setEnabled(False)
        if not self.im_info.no_t and self.num_frames_max is not None:
            if self.single:
                self.nellie.processor.time_input.setEnabled(True)
                self.nellie.processor.time_input.setRange(1, self.num_frames_max)
                self.nellie.processor.time_input.setValue(self.num_frames_max)
            else:
                self.nellie.batch_mode.time_input.setEnabled(True)
                self.nellie.batch_mode.time_input.setRange(1, self.num_frames_max)
                self.nellie.batch_mode.time_input.setValue(self.num_frames_max)
        else:
            self.nellie.processor.time_input.setRange(self.num_frames_max, self.num_frames_max)
            self.nellie.processor.time_input.setValue(self.num_frames_max)
            self.nellie.batch_mode.time_input.setRange(self.num_frames_max, self.num_frames_max)
            self.nellie.batch_mode.time_input.setValue(self.num_frames_max)

    def enable_channel_options(self):
        if not self.im_info.no_c and self.num_channels > 1:
            self.nellie.processor.channel_input.setEnabled(True)
            self.nellie.processor.channel_input.setRange(0, self.num_channels - 1)
            self.nellie.processor.channel_input.setValue(self.nellie.processor.channel_input.value())
        else:
            self.nellie.processor.channel_input.setEnabled(False)
            self.nellie.processor.channel_input.setRange(0, 0)
            self.nellie.processor.channel_input.setValue(0)

    def post_file_selection(self):
        self.set_max_frames()
        self.set_num_channels()

        # set valid tabs and buttons to be enabled
        self.check_analysis_valid()
        self.disable_file_selection()
        self.enable_channel_options()
        self.enable_time_options()
        if not self.im_info_valid:
            return
        if self.single:
            self.nellie.setTabEnabled(self.nellie.processor_tab, True)
            self.nellie.setTabEnabled(self.nellie.visualizer_tab, True)
            self.nellie.setTabEnabled(self.nellie.batch_tab, False)
        else:
            self.nellie.setTabEnabled(self.nellie.processor_tab, False)
            self.nellie.setTabEnabled(self.nellie.visualizer_tab, False)
            self.nellie.setTabEnabled(self.nellie.analysis_tab, False)
            self.nellie.setTabEnabled(self.nellie.batch_tab, True)
        self.nellie.file_ready()

    def select_folder(self):
        filepath = QFileDialog.getExistingDirectory(self, "Select folder")
        self.validate_path(filepath)
        if self.filepath is None:
            return
        self.single = False

        self.initialize_folder(filepath)
        filename = os.path.basename(self.filepath)
        show_info(f"Selected folder: {filename}")
        self.filepath_text.setText(f"Selected folder: {filename}")

        self.post_file_selection()

    def validate_path(self, filepath):
        if not filepath:
            show_info("Invalid selection.")
            return None
        self.filepath = filepath

    def select_filepath(self):
        filepath, _ = QFileDialog.getOpenFileName(self, "Select file")
        self.validate_path(filepath)
        if self.filepath is None:
            return
        self.single = True

        self.initialize_single_file()
        filename = os.path.basename(self.filepath)
        show_info(f"Selected file: {filename}")
        self.filepath_text.setText(f"Selected file: {filename}")

        self.post_file_selection()

    def reset(self):
        # switch to nellie's file_select tab
        self.nellie.reset()


if __name__ == "__main__":
    import napari
    viewer = napari.Viewer()
    napari.run()



import numpy as np
import skimage.measure
import skimage.morphology as morph
from scipy.spatial import cKDTree

from nellie import xp, ndi, logger, device_type
from nellie.im_info.im_info import ImInfo
from nellie.utils.general import get_reshaped_image
from nellie.utils.gpu_functions import triangle_threshold, otsu_threshold


class Network:
    def __init__(self, im_info: ImInfo, num_t=None,
                 min_radius_um=0.20, max_radius_um=1, clean_skel=None):
        self.im_info = im_info
        self.num_t = num_t
        if num_t is None and not self.im_info.no_t:
            self.num_t = im_info.shape[im_info.axes.index('T')]
        if not self.im_info.no_z:
            if clean_skel is None:
                clean_skel = False
            self.z_ratio = self.im_info.dim_sizes['Z'] / self.im_info.dim_sizes['X']
        # either (roughly) diffraction limit, or pixel size, whichever is larger
        self.min_radius_um = max(min_radius_um, self.im_info.dim_sizes['X'])
        self.max_radius_um = max_radius_um

        self.min_radius_px = self.min_radius_um / self.im_info.dim_sizes['X']
        self.max_radius_px = self.max_radius_um / self.im_info.dim_sizes['X']

        if self.im_info.no_z:
            self.scaling = (im_info.dim_sizes['Y'], im_info.dim_sizes['X'])
        else:
            self.scaling = (im_info.dim_sizes['Z'], im_info.dim_sizes['Y'], im_info.dim_sizes['X'])

        self.shape = ()

        self.im_memmap = None
        self.im_frangi_memmap = None
        self.label_memmap = None
        self.network_memmap = None
        self.pixel_class_memmap = None
        self.skel_memmap = None
        self.skel_relabelled_memmap = None

        self.clean_skel = True if clean_skel is None else clean_skel

        self.sigmas = None

        self.debug = None

    def _remove_connected_label_pixels(self, skel_labels):
        if device_type == 'cuda':
            skel_labels = skel_labels.get()

        if self.im_info.no_z:
            height, width = skel_labels.shape
        else:
            depth, height, width = skel_labels.shape

        true_coords = np.argwhere(skel_labels > 0)

        pixels_to_delete = []
        for coord in true_coords:
            if self.im_info.no_z:
                y, x = coord
            else:
                z, y, x = coord

            if not self.im_info.no_z:
                if z == 0 or z == depth - 1:
                    continue
            if y == 0 or y == height - 1 or x == 0 or x == width - 1:
                continue  # skip boundary voxels

            # extract 3x3x3 neighborhood
            if self.im_info.no_z:
                label_neighborhood = skel_labels[y - 1:y + 2, x - 1:x + 2]
            else:
                label_neighborhood = skel_labels[z - 1:z + 2, y - 1:y + 2, x - 1:x + 2]

            # get labels of set voxels in the neighborhood
            labels_in_neighborhood = label_neighborhood[label_neighborhood > 0]

            if len(set(labels_in_neighborhood.tolist())) > 1:
                if self.im_info.no_z:
                    pixels_to_delete.append((y, x))
                else:
                    pixels_to_delete.append((z, y, x))

        if self.im_info.no_z:
            for y, x in pixels_to_delete:
                skel_labels[y, x] = 0
        else:
            for z, y, x in pixels_to_delete:
                skel_labels[z, y, x] = 0

        return xp.array(skel_labels)

    def _add_missing_skeleton_labels(self, skel_frame, label_frame, frangi_frame, thresh):
        logger.debug('Adding missing skeleton labels.')
        gpu_frame = xp.array(label_frame)
        # identify unique labels and find missing ones
        unique_labels = xp.unique(gpu_frame)
        unique_skel_labels = xp.unique(skel_frame)

        missing_labels = set(unique_labels.tolist()) - set(unique_skel_labels.tolist())

        # for each missing label, find the centroid and mark it in the skeleton
        for label in missing_labels:
            if label == 0:  # ignore bg label
                continue

            label_coords = xp.argwhere(gpu_frame == label)
            label_intensities = frangi_frame[tuple(label_coords.T)]
            # centroid is where label_intensities is maximal
            centroid = label_coords[xp.argmax(label_intensities)]

            skel_frame[tuple(centroid)] = label

        return skel_frame

    def _skeletonize(self, label_frame, frangi_frame):
        cpu_frame = np.array(label_frame)
        gpu_frame = xp.array(label_frame)

        skel = xp.array(morph.skeletonize(cpu_frame > 0).astype('bool'))

        if self.clean_skel:
            masked_frangi = ndi.median_filter(frangi_frame, size=3) * (gpu_frame > 0)  # * skel
            thresh_otsu, _ = otsu_threshold(xp.log10(masked_frangi[masked_frangi > 0]))
            thresh_otsu = 10 ** thresh_otsu
            thresh_tri = triangle_threshold(xp.log10(masked_frangi[masked_frangi > 0]))
            thresh_tri = 10 ** thresh_tri
            thresh = min(thresh_otsu, thresh_tri)
            cleaned_skel = (masked_frangi > thresh) * skel

            skel_labels = gpu_frame * cleaned_skel
            label_sizes = xp.bincount(skel_labels.ravel())

            above_threshold = label_sizes > 1

            mask = xp.zeros_like(skel_labels, dtype=bool)
            mask[above_threshold[skel_labels]] = True
            mask[skel_labels == 0] = False

            skel_labels = gpu_frame * mask
        else:
            skel_labels = gpu_frame * skel
            thresh = 0

        return skel_labels, thresh

    def _get_sigma_vec(self, sigma):
        if self.im_info.no_z:
            sigma_vec = (sigma, sigma)
        else:
            sigma_vec = (sigma / self.z_ratio, sigma, sigma)
        return sigma_vec

    def _set_default_sigmas(self):
        logger.debug('Setting to sigma values.')
        min_sigma_step_size = 0.2
        num_sigma = 5

        self.sigma_min = self.min_radius_px / 2
        self.sigma_max = self.max_radius_px / 3

        sigma_step_size_calculated = (self.sigma_max - self.sigma_min) / num_sigma
        sigma_step_size = max(min_sigma_step_size, sigma_step_size_calculated)  # Avoid taking too small of steps.

        self.sigmas = list(xp.arange(self.sigma_min, self.sigma_max, sigma_step_size))
        logger.debug(f'Calculated sigma step size = {sigma_step_size_calculated}. Sigmas = {self.sigmas}')

    def _relabel_objects(self, branch_skel_labels, label_frame):
        if self.im_info.no_z:
            structure = xp.ones((3, 3))
        else:
            structure = xp.ones((3, 3, 3))
        # here, skel frame should be the branch labeled frame
        relabelled_labels = branch_skel_labels.copy()
        skel_mask = xp.array(branch_skel_labels > 0).astype('uint8')
        label_mask = xp.array(label_frame > 0).astype('uint8')
        skel_border = (ndi.binary_dilation(skel_mask, iterations=1, structure=structure) ^ skel_mask) * label_mask
        skel_label_mask = (branch_skel_labels > 0)

        if device_type == 'cuda':
            skel_label_mask = skel_label_mask.get()

        vox_matched = np.argwhere(skel_label_mask)

        if device_type == 'cuda':
            vox_next_unmatched = np.argwhere(skel_border.get())
        else:
            vox_next_unmatched = np.argwhere(skel_border)

        unmatched_diff = np.inf
        while True:
            num_unmatched = len(vox_next_unmatched)
            if num_unmatched == 0:
                break
            tree = cKDTree(vox_matched * self.scaling)
            dists, idxs = tree.query(vox_next_unmatched * self.scaling, k=1, workers=-1)
            # remove any matches that are too far away
            max_dist = 2 * np.min(self.scaling)  # sqrt 3 * max scaling
            unmatched_matches = np.array(
                [[vox_matched[idx], vox_next_unmatched[i]] for i, idx in enumerate(idxs) if dists[i] < max_dist]
            )
            if len(unmatched_matches) == 0:
                break
            matched_labels = branch_skel_labels[tuple(np.transpose(unmatched_matches[:, 0]))]
            relabelled_labels[tuple(np.transpose(unmatched_matches[:, 1]))] = matched_labels
            branch_skel_labels = relabelled_labels.copy()
            relabelled_labels_mask = relabelled_labels > 0

            if device_type == 'cuda':
                relabelled_labels_mask_cpu = relabelled_labels_mask.get()
            else:
                relabelled_labels_mask_cpu = relabelled_labels_mask

            vox_matched = np.argwhere(relabelled_labels_mask_cpu)
            relabelled_mask = relabelled_labels_mask.astype('uint8')
            # add unmatched matches to coords_matched
            skel_border = (ndi.binary_dilation(relabelled_mask, iterations=1,
                                               structure=structure) - relabelled_mask) * label_mask

            if device_type == 'cuda':
                vox_next_unmatched = np.argwhere(skel_border.get())
            else:
                vox_next_unmatched = np.argwhere(skel_border)

            new_num_unmatched = len(vox_next_unmatched)
            unmatched_diff_temp = abs(num_unmatched - new_num_unmatched)
            if unmatched_diff_temp == unmatched_diff:
                break
            unmatched_diff = unmatched_diff_temp
            logger.debug(f'Reassigned {unmatched_diff}/{num_unmatched} unassigned voxels. '
                         f'{new_num_unmatched} remain.')

        return relabelled_labels

    def _local_max_peak(self, frame, mask):
        lapofg = xp.empty(((len(self.sigmas),) + frame.shape), dtype=float)
        for i, s in enumerate(self.sigmas):
            sigma_vec = self._get_sigma_vec(s)
            current_lapofg = -ndi.gaussian_laplace(frame, sigma_vec) * xp.mean(s) ** 2
            current_lapofg = current_lapofg * mask
            current_lapofg[current_lapofg < 0] = 0
            lapofg[i] = current_lapofg

        filt_footprint = xp.ones((3,) * (frame.ndim + 1))
        max_filt = ndi.maximum_filter(lapofg, footprint=filt_footprint, mode='nearest')
        peaks = xp.empty(lapofg.shape, dtype=bool)
        max_filt_mask = mask
        for filt_slice, max_filt_slice in enumerate(max_filt):
            peaks[filt_slice] = (xp.asarray(lapofg[filt_slice]) == xp.asarray(max_filt_slice)) * max_filt_mask
        # get the coordinates of all true pixels in peaks
        coords = xp.max(peaks, axis=0)
        coords_3d = xp.argwhere(coords)
        peak_im = xp.zeros_like(frame)
        peak_im[tuple(coords_3d.T)] = 1
        return coords_3d

    def _get_pixel_class(self, skel):
        skel_mask = xp.array(skel > 0).astype('uint8')
        if self.im_info.no_z:
            weights = xp.ones((3, 3))
        else:
            weights = xp.ones((3, 3, 3))
        skel_mask_sum = ndi.convolve(skel_mask, weights=weights, mode='constant', cval=0) * skel_mask
        skel_mask_sum[skel_mask_sum > 4] = 4
        return skel_mask_sum

    def _get_t(self):
        if self.num_t is None:
            if self.im_info.no_t:
                self.num_t = 1
            else:
                self.num_t = self.im_info.shape[self.im_info.axes.index('T')]
        else:
            return

    def _allocate_memory(self):
        logger.debug('Allocating memory for skeletonization.')
        label_memmap = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_instance_label'])  # , read_type='r+')
        self.label_memmap = get_reshaped_image(label_memmap, self.num_t, self.im_info)

        im_memmap = self.im_info.get_im_memmap(self.im_info.im_path)
        self.im_memmap = get_reshaped_image(im_memmap, self.num_t, self.im_info)

        im_frangi_memmap = self.im_info.get_im_memmap(self.im_info.pipeline_paths['im_frangi'])
        self.im_frangi_memmap = get_reshaped_image(im_frangi_memmap, self.num_t, self.im_info)
        self.shape = self.label_memmap.shape

        im_skel_path = self.im_info.pipeline_paths['im_skel']
        self.skel_memmap = self.im_info.allocate_memory(im_skel_path, shape=self.shape,
                                                        dtype='uint16',
                                                        description='skeleton image',
                                                        return_memmap=True)

        im_pixel_class = self.im_info.pipeline_paths['im_pixel_class']
        self.pixel_class_memmap = self.im_info.allocate_memory(im_pixel_class, shape=self.shape,
                                                               dtype='uint8',
                                                               description='pixel class image',
                                                               return_memmap=True)

        im_skel_relabelled = self.im_info.pipeline_paths['im_skel_relabelled']
        self.skel_relabelled_memmap = self.im_info.allocate_memory(im_skel_relabelled, shape=self.shape,
                                                                   dtype='uint32',
                                                                   description='skeleton relabelled image',
                                                                   return_memmap=True)

    def _get_branch_skel_labels(self, pixel_class):
        # get the labels of the skeleton pixels that are not junctions or background
        non_junctions = pixel_class > 0
        non_junctions = non_junctions * (pixel_class != 4)
        if self.im_info.no_z:
            structure = xp.ones((3, 3))
        else:
            structure = xp.ones((3, 3, 3))
        non_junction_labels, _ = ndi.label(non_junctions, structure=structure)
        return non_junction_labels

    def _run_frame(self, t):
        logger.info(f'Running network analysis, volume {t}/{self.num_t - 1}')
        label_frame = self.label_memmap[t]
        frangi_frame = xp.array(self.im_frangi_memmap[t])
        skel_frame, thresh = self._skeletonize(label_frame, frangi_frame)
        skel = self._remove_connected_label_pixels(skel_frame)
        skel = self._add_missing_skeleton_labels(skel, label_frame, frangi_frame, thresh)
        if device_type == 'cuda':
            skel = skel.get()

        skel_pre = (skel > 0) * label_frame
        pixel_class = self._get_pixel_class(skel_pre)
        branch_skel_labels = self._get_branch_skel_labels(pixel_class)
        branch_labels = self._relabel_objects(branch_skel_labels, label_frame)
        return branch_skel_labels, pixel_class, branch_labels

    def _clean_junctions(self, pixel_class):
        junctions = pixel_class == 4
        junction_labels = skimage.measure.label(junctions)
        junction_objects = skimage.measure.regionprops(junction_labels)
        junction_centroids = [obj.centroid for obj in junction_objects]
        for junction_num, junction in enumerate(junction_objects):
            # use ckd tree to find closest junction coord to junction centroid
            if len(junction.coords) < 2:
                continue
            junction_tree = cKDTree(junction.coords)
            _, nearest_junction_indices = junction_tree.query(junction_centroids[junction_num], k=1, workers=-1)
            # remove the nearest junction coord from the junction
            junction_coords = junction.coords.tolist()
            junction_coords.pop(nearest_junction_indices)
            pixel_class[tuple(np.array(junction_coords).T)] = 3
        return pixel_class

    def _run_networking(self):
        for t in range(self.num_t):
            skel, pixel_class, skel_relabelled_memmap = self._run_frame(t)
            if self.im_info.no_t:
                if device_type == 'cuda':
                    self.skel_memmap[:] = skel[:].get()
                    self.pixel_class_memmap[:] = pixel_class[:].get()
                    self.skel_relabelled_memmap[:] = skel_relabelled_memmap[:].get()
                else:
                    self.skel_memmap[:] = skel[:]
                    self.pixel_class_memmap[:] = pixel_class[:]
                    self.skel_relabelled_memmap[:] = skel_relabelled_memmap[:]
            else:
                if device_type == 'cuda':
                    self.skel_memmap[t] = skel.get()
                    self.pixel_class_memmap[t] = pixel_class.get()
                    self.skel_relabelled_memmap[t] = skel_relabelled_memmap.get()
                else:
                    self.skel_memmap[t] = skel
                    self.pixel_class_memmap[t] = pixel_class
                    self.skel_relabelled_memmap[t] = skel_relabelled_memmap

    def run(self):
        self._get_t()
        self._allocate_memory()
        self._run_networking()


if __name__ == "__main__":
    im_path = r"D:\test_files\nelly_tests\deskewed-2023-07-13_14-58-28_000_wt_0_acquire.ome.tif"
    im_info = ImInfo(im_path)
    skel = Network(im_info, num_t=3)
    skel.run()



from qtpy.QtWidgets import QLabel, QVBoxLayout, QWidget, QPushButton, QMessageBox
from qtpy.QtGui import QPixmap, QFont
from qtpy.QtCore import Qt
import napari
import os
import datetime
import tifffile
from napari.utils.notifications import show_info


class Home(QWidget):
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', nellie, parent=None):
        super().__init__(parent)
        self.nellie = nellie
        self.viewer = napari_viewer

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        # Logo
        logo_path = os.path.join(os.path.dirname(__file__), 'logo.png')
        logo_label = QLabel(self)
        pixmap = QPixmap(logo_path)
        logo_label.setPixmap(
            pixmap.scaled(300, 300, Qt.KeepAspectRatio))#, Qt.SmoothTransformation))  # Adjust size as needed
        logo_label.setAlignment(Qt.AlignCenter)
        self.layout.addWidget(logo_label)

        # Title
        title = QLabel("Nellie")
        title.setFont(QFont("Arial", 24, QFont.Bold))
        title.setAlignment(Qt.AlignCenter)
        self.layout.addWidget(title)

        # Subtitle
        subtitle = QLabel("Automated organelle segmentation, tracking, and hierarchical feature extraction in 2D/3D live-cell microscopy.")
        subtitle.setFont(QFont("Arial", 16))
        subtitle.setAlignment(Qt.AlignCenter)
        subtitle.setWordWrap(True)
        self.layout.addWidget(subtitle)

        self.layout.addWidget(QLabel("\n"))  # Add a bit of space

        # todo link to paper
        github_link = QLabel("<a href='https://arxiv.org/abs/2403.13214'>Cite our paper!</a>")
        github_link.setOpenExternalLinks(True)
        github_link.setAlignment(Qt.AlignCenter)
        self.layout.addWidget(github_link)

        # Link to GitHub
        github_link = QLabel("<a href='https://github.com/aelefebv/nellie'>Visit Nellie's GitHub Page!</a>")
        github_link.setOpenExternalLinks(True)
        github_link.setAlignment(Qt.AlignCenter)
        self.layout.addWidget(github_link)

        # screenshot button
        self.screenshot_button = QPushButton(text="Easy screenshot:\n[Ctrl/Cmd-Shift-E]")
        self.screenshot_button.clicked.connect(self.screenshot)
        self.screenshot_button.setEnabled(True)
        self.viewer.bind_key('Ctrl-Shift-E', self.screenshot, overwrite=True)

        self.layout.addWidget(self.screenshot_button, alignment=Qt.AlignCenter)

    def screenshot(self, event=None):
        if self.nellie.im_info is None:
            show_info("No file selected, cannot take screenshot")
            return

        # easy no prompt screenshot
        dt = datetime.datetime.now()  # year, month, day, hour, minute, second, millisecond up to 3 digits
        dt = dt.strftime("%Y%m%d_%H%M%S%f")[:-3]

        screenshot_folder = self.nellie.im_info.screenshot_dir
        if not os.path.exists(screenshot_folder):
            os.makedirs(screenshot_folder)

        im_name = f'{dt}-{self.nellie.im_info.basename_no_ext}.png'
        file_path = os.path.join(screenshot_folder, im_name)

        # Take screenshot
        screenshot = self.viewer.screenshot(canvas_only=True)

        # Save the screenshot
        try:
            # save as png to file_path using tifffile
            tifffile.imwrite(file_path, screenshot)
            show_info(f"Screenshot saved to {file_path}")
        except Exception as e:
            QMessageBox.warning(None, "Error", f"Failed to save screenshot: {str(e)}")
            raise e


if __name__ == "__main__":
    import napari
    viewer = napari.Viewer()
    napari.run()



import numpy as np
import torch


def flip(input_array, dims):
    if not isinstance(input_array, torch.Tensor):
        input_array = torch.tensor(input_array)
    return torch.flip(input_array, dims=dims)


def array(input_array, dtype=None, device=None):
    # Convert input to a PyTorch tensor
    if isinstance(input_array, torch.Tensor):
        result = input_array.clone()
    elif isinstance(input_array, np.ndarray):
        result = torch.from_numpy(input_array)
    # if its a list of tensors, convert to a single tensor
    elif isinstance(input_array, list):
        if isinstance(input_array[0], list):
            tensor_stack = []
            for tensor in input_array:
                tensor_stack.append(tensor[0])
            input_array = tensor_stack
        result = torch.stack(input_array)
    else:
        result = torch.tensor(input_array)

    # Handle dtype
    if dtype is not None:
        result = result.to(dtype)

    # Handle device
    if device is not None:
        result = result.to(device)

    return result


def to_tensor(input_array):
    if not isinstance(input_array, torch.Tensor):
        if isinstance(input_array, np.ndarray):
            try:
                from_numpy = torch.from_numpy(input_array)
            except:
                from_numpy = torch.from_numpy(input_array.astype(np.float32))
                # logger.warning(f'failed to convert numpy array to tensor, converted to float32\n{input_array}')
            return from_numpy
        elif isinstance(input_array, list):
            if isinstance(input_array[0], list):
                tensor_stack = []
                for tensor in input_array:
                    tensor_stack.append(tensor[0])
                return torch.stack(tensor_stack)
            elif isinstance(input_array[0], torch.Tensor):
                return input_array
        else:
            return torch.tensor(input_array)
    return input_array


def var(input_array):
    # make sure input is a float tensor
    input_array = to_tensor(input_array).float()
    return torch.var(input_array, unbiased=False)


def histogram(input_array, bins, range):
    input_array = to_tensor(input_array).float()
    hist = torch.histc(input_array, bins=bins, min=range[0], max=range[1])
    bin_edges = torch.linspace(range[0], range[1], steps=bins + 1)
    return hist, bin_edges


def sum(input_array):
    input_array = to_tensor(input_array)
    return torch.sum(input_array)


def cumsum(input_array):
    input_array = to_tensor(input_array)
    return torch.cumsum(input_array, dim=0)


def argmax(input_array):
    input_array = to_tensor(input_array)
    return torch.argmax(input_array, dim=0)


def flatnonzero(input_array):
    input_array = to_tensor(input_array)
    return torch.nonzero(input_array.flatten(), as_tuple=False).flatten()


def arange(start, end=None, step=1):
    # Note: arange typically doesn't need conversion to tensor for its arguments
    if end is None:
        return torch.arange(start)
    else:
        return torch.arange(start, end, step)


def sqrt(input_array):
    input_array = to_tensor(input_array)
    return torch.sqrt(input_array)


def zeros_like(input_array, dtype=None):
    input_array = to_tensor(input_array)
    if dtype == 'float64':
        dtype = torch.float64
    return torch.zeros_like(input_array, dtype=dtype)


def ones_like(input_array, dtype=None):
    input_array = to_tensor(input_array)
    if dtype == 'bool':
        dtype = torch.bool
    return torch.ones_like(input_array, dtype=dtype)


def ones(shape, dtype=None):
    if dtype == 'bool':
        dtype = torch.bool
    return torch.ones(shape, dtype=dtype)


def asarray(input_array, dtype=None):
    tensor_out = to_tensor(input_array)
    if dtype is not None:
        if dtype == 'double':
            dtype = torch.float64
        tensor_out = tensor_out.to(dtype)
    return tensor_out


def asnumpy(input_array):
    return input_array.cpu().numpy()


def gradient(input_array, axis=None):
    input_array = to_tensor(input_array)
    grad = list(torch.gradient(input_array, dim=axis))
    return grad


def max(input_array):
    input_array = to_tensor(input_array)
    return torch.max(input_array)


def min(input_array):
    input_array = to_tensor(input_array)
    return torch.min(input_array)


def abs(input_array):
    input_array = to_tensor(input_array)
    return torch.abs(input_array)


# def linalg(input_array, axis=None):
#     input_array = to_tensor(input_array)
#     return torch.linalg.norm(input_array, dim=axis)

def isinf(input_array):
    input_array = to_tensor(input_array)
    return torch.isinf(input_array)


def concatenate(input_array, axis=0):
    input_array = to_tensor(input_array)
    return torch.cat(input_array, dim=axis)


def argsort(input_array, axis=None):
    input_array = to_tensor(input_array)
    return torch.argsort(input_array, dim=axis)


def take_along_axis(input_array, indices, axis=None):
    input_array = to_tensor(input_array)
    return torch.take_along_dim(input_array, indices, dim=axis)


def exp(input_array):
    input_array = to_tensor(input_array)
    return torch.exp(input_array)


def nan_to_num(input_array, copy=True, nan=0.0):
    input_array = to_tensor(input_array)
    return torch.nan_to_num(input_array, nan=nan)


def where(condition, x, y):
    condition = to_tensor(condition)
    x = to_tensor(x)
    y = to_tensor(y)
    return torch.where(condition, x, y)


def percentile(input_array, q, axis=None):
    input_array = to_tensor(input_array)
    q /= 100
    return torch.quantile(input_array, q, dim=axis)


def inf():
    return torch.tensor(float('inf'))


def log10(input_array):
    input_array = to_tensor(input_array)
    return torch.log10(input_array)


def mean(input_array, axis=None):
    input_array = to_tensor(input_array)
    return torch.mean(input_array, dim=axis)



import os

import tifffile
import nd2
import ome_types
from nellie import logger
import numpy as np
from typing import Union, Type


class ImInfo:
    def __init__(self, im_path: str,
                 output_dirpath: str = None,
                 screenshot_dirpath: str = None,
                 ch: int = None,
                 dim_sizes: dict = None,
                 dimension_order: str = '',
                 output_suffix: str = '',
                 ):
        self.im_path = im_path
        self.ch = ch or 0

        self.dim_sizes = dim_sizes

        self.axes = dimension_order
        self.axes_valid = True

        self.output_suffix = output_suffix

        self.extension = os.path.splitext(self.im_path)[1]
        self.basename = os.path.basename(self.im_path)
        self.basename_no_ext = os.path.splitext(self.basename)[0]
        self.dirname = os.path.basename(os.path.dirname(self.im_path))
        self.input_dir = os.path.dirname(self.im_path)
        self.output_dir = output_dirpath or os.path.join(self.input_dir, 'nellie_output')
        self.output_dir = self.output_dir + self.output_suffix
        self.screenshot_dir = screenshot_dirpath or os.path.join(self.output_dir, 'screenshots')
        self.screenshot_dir = self.screenshot_dir + self.output_suffix
        self.graph_dir = os.path.join(self.output_dir, 'graphs')
        self.graph_dir = self.graph_dir + self.output_suffix

        self.shape = ()
        self.metadata = None
        self.metadata_type = None
        self._load_metadata()

        if self.dim_sizes is None:
            self._set_dim_sizes()

        self.no_z = True
        self.no_t = True
        self.no_c = True
        self._check_axes()

        self._create_output_dir()

        self.pipeline_paths = {}
        self._create_output_paths()

        self._check_memmappable()

    def _create_output_paths(self):
        self.create_output_path('im_frangi')
        self.create_output_path('im_instance_label')
        self.create_output_path('im_skel')
        self.create_output_path('im_skel_relabelled')
        self.create_output_path('im_pixel_class')
        self.create_output_path('im_marker')
        self.create_output_path('im_distance')
        self.create_output_path('im_border')
        self.create_output_path('flow_vector_array', ext='.npy')
        self.create_output_path('voxel_matches', ext='.npy')
        self.create_output_path('im_branch_label_reassigned')
        self.create_output_path('im_obj_label_reassigned')
        self.create_output_path('features_voxels', ext='.csv')
        self.create_output_path('features_nodes', ext='.csv')
        self.create_output_path('features_branches', ext='.csv')
        self.create_output_path('features_components', ext='.csv')
        self.create_output_path('features_image', ext='.csv')
        self.create_output_path('adjacency_maps', ext='.pkl')

    def _create_output_dir(self):
        if not os.path.isdir(self.output_dir):
            os.mkdir(self.output_dir)

    def _check_memmappable(self):
        try:
            tifffile.memmap(self.im_path, mode='r')
        except (ValueError, tifffile.tifffile.TiffFileError):
            logger.warning(f'Could not create memmap for {self.im_path}. Loading into memory instead.')
            self._get_ome_tif()
            self.im_path = self.pipeline_paths['ome']
            self.extension = '.ome.tif'

    def _get_ome_tif(self):
        ome_path = self.create_output_path('ome')
        if os.path.isfile(ome_path):
            return
        if self.extension == '.nd2':
            data = nd2.imread(self.im_path)
        elif self.extension == '.tif' or self.extension == '.tiff':
            # get only self.ch
            data = tifffile.imread(self.im_path)
        else:
            logger.error(f'Filetype {self.extension} not supported. Please convert to .nd2 or .tif.')
            raise ValueError
        if not self.no_c:
            data = np.take(data, self.ch, axis=self.axes.index('C'))
        self.allocate_memory(ome_path, data=data)

    def _check_axes(self):
        if 'Z' in self.axes and self.shape[self.axes.index('Z')] > 1:
            self.no_z = False
        if 'T' in self.axes and self.shape[self.axes.index('T')] > 1:
            self.no_t = False
        if 'C' in self.axes and self.shape[self.axes.index('C')] > 1:
            self.no_c = False

    def _get_tif_tag_metadata(self, metadata):
        tag_names = {tag_value.name: tag_code for tag_code, tag_value in metadata.items()}

        if 'XResolution' in tag_names:
            self.dim_sizes['X'] = metadata[tag_names['XResolution']].value[1] \
                                  / metadata[tag_names['XResolution']].value[0]
        else:
            logger.warning('No XResolution tag found.')
            self.dim_sizes['X'] = None

        if 'YResolution' in tag_names:
            self.dim_sizes['Y'] = metadata[tag_names['YResolution']].value[1] \
                                  / metadata[tag_names['YResolution']].value[0]
        else:
            logger.warning('No YResolution tag found.')
            self.dim_sizes['Y'] = None

        if 'ResolutionUnit' in tag_names:
            if metadata[tag_names['ResolutionUnit']].value == tifffile.TIFF.RESUNIT.CENTIMETER:
                self.dim_sizes['X'] *= 1E-2 * 1E6
                self.dim_sizes['Y'] *= 1E-2 * 1E6
        if 'Z' in self.axes:
            if 'ZResolution' in tag_names:
                self.dim_sizes['Z'] = 1 / metadata[tag_names['ZResolution']].value[0]
            else:
                logger.warning('No ZResolution tag found.')
                self.dim_sizes['Z'] = None
        else:
            logger.warning('No ZResolution tag found.')
            self.dim_sizes['Z'] = None

        if 'T' in self.axes:
            if 'FrameRate' in tag_names:
                self.dim_sizes['T'] = 1 / metadata[tag_names['FrameRate']].value[0]
        else:
            logger.warning('No FrameRate tag found.')
            self.dim_sizes['T'] = None

    def _get_imagej_metadata(self, metadata):
        if 'physicalsizex' in metadata:
            self.dim_sizes['X'] = metadata['physicalsizex']
        if 'physicalsizey' in metadata:
            self.dim_sizes['Y'] = metadata['physicalsizey']
        if 'spacing' in metadata:
            self.dim_sizes['Z'] = metadata['spacing']
        if 'finterval' in metadata:
            self.dim_sizes['T'] = metadata['finterval']

    def _set_dim_sizes(self):
        if self.dim_sizes is not None:
            return
        try:
            self.dim_sizes = {'X': None, 'Y': None, 'Z': None, 'T': None}
            if self.metadata_type == 'imagej':
                self._get_imagej_metadata(self.metadata)

            elif self.metadata_type == 'ome':
                self.dim_sizes['X'] = self.metadata.images[0].pixels.physical_size_x
                self.dim_sizes['Y'] = self.metadata.images[0].pixels.physical_size_y
                self.dim_sizes['Z'] = self.metadata.images[0].pixels.physical_size_z
                self.dim_sizes['T'] = self.metadata.images[0].pixels.time_increment
            elif self.metadata_type == 'imagej_tif_tags':
                self._get_tif_tag_metadata(self.metadata[1])
                self._get_imagej_metadata(self.metadata[0])
                logger.warning(f'File is an ImageJ type, but has weird metadata, '
                               f'found dimension sizes: {self.dim_sizes}')

            elif self.metadata_type is None:
                self._get_tif_tag_metadata(self.metadata)
                logger.warning(f'File is not an ImageJ or OME type, found dimension sizes: {self.dim_sizes}')

            elif self.metadata_type == 'nd2':
                try:
                    timestamps = self.metadata.recorded_data['Time [s]']
                    self.dim_sizes['T'] = timestamps[-1] / len(timestamps)
                except KeyError:
                    logger.warning('No time data found in ND2 file.')
                    self.dim_sizes['T'] = None
                self.dim_sizes['X'] = self.metadata.volume.axesCalibration[0]
                self.dim_sizes['Y'] = self.metadata.volume.axesCalibration[1]
                self.dim_sizes['Z'] = self.metadata.volume.axesCalibration[2]
            self.metadata = None
            if self.dim_sizes['X'] is None:
                logger.error('No X dimension found.')
                raise ValueError
            if self.dim_sizes['X'] != self.dim_sizes['Y']:
                logger.warning('X and Y dimensions do not match. Non-square pixels not supported, '
                               'so unexpected results and wrong measurements will occur.')
            if 'T' in self.axes:
                if self.dim_sizes['T'] is None:
                    logger.warning('No FrameRate tag found.')
                    self.dim_sizes['T'] = None
        except Exception as e:
            logger.error(f"Error loading metadata for image {self.im_path}: {str(e)}")
            self.metadata = {}
            self.dim_sizes = {}
            raise e

    def _load_tif(self):
        with tifffile.TiffFile(self.im_path) as tif:
            if tif.is_ome or tif.ome_metadata is not None:
                ome_xml = tifffile.tiffcomment(self.im_path)
                ome = ome_types.from_xml(ome_xml, parser="lxml")
                self.metadata = ome
                self.metadata_type = 'ome'
            elif tif.is_imagej:
                self.metadata = tif.imagej_metadata
                self.metadata_type = 'imagej'
                if 'physicalsizex' not in self.metadata:
                    self.metadata_type = 'imagej_tif_tags'
                    self.metadata = [self.metadata, tif.pages[0].tags._dict]
            else:
                self.metadata = tif.pages[0].tags._dict
                self.metadata_type = None
                if 'C' in self.axes:
                    logger.error('Multichannel TIFF files must have ImageJ or OME metadata. Resubmit the'
                                 'file either with correct metadata or the single channel of interest.')
            if self.axes == '':
                self.axes = tif.series[0].axes
            self.shape = tif.series[0].shape
            if len(self.axes) != len(self.shape):
                logger.error(f"Dimension order {self.axes} does not match the number of dimensions in the image "
                             f"({len(self.shape)}).")
                # raise error
                raise ValueError

    def _load_nd2(self):
        with nd2.ND2File(self.im_path) as nd2_file:
            self.metadata = nd2_file.metadata.channels[self.ch]
            self.metadata.recorded_data = nd2_file.events(orient='list')
            self.metadata_type = 'nd2'
            if self.axes == '':
                self.axes = ''.join(nd2_file.sizes.keys())
            self.shape = tuple(nd2_file.sizes.values())
            if len(self.axes) != len(self.shape):
                logger.error(f"Dimension order {self.axes} does not match the number of dimensions in the image "
                             f"({len(self.shape)}).")
                raise ValueError

    def _load_metadata(self):
        try:
            if self.im_path.endswith('.tif') or self.im_path.endswith('.tiff'):
                self._load_tif()
            elif self.im_path.endswith('.nd2'):
                self._load_nd2()

        except Exception as e:
            logger.error(f"Error loading file {self.im_path}")
            raise e

        if 'Q' in self.axes:
            # change to T
            self.axes = self.axes.replace('Q', 'T')

        accepted_axes = ['TZYX', 'TYX', 'TZCYX', 'TCYX', 'TCZYX', 'ZYX', 'YX', 'CYX', 'CZYX', 'ZCYX']
        if self.axes not in accepted_axes:
            # todo, have user optionally specify axes
            logger.warning(f"File dimension order is in unknown order {self.axes} with {len(self.shape)} dimensions. \n"
                           f"Please specify the order of the dimensions in the run with the "
                           f"'dimension_order' parameter. \n"
                           f"Accepted dimensions are: {accepted_axes}.")
            self.axes_valid = False
            # raise ValueError

    def _ensure_t(self, data=None):
        if 'T' not in self.axes:
            self.axes = 'T' + self.axes
        if len(self.axes) != len(self.shape):
            self.shape = (1,) + self.shape
            if data is not None:
                data = np.expand_dims(data, axis=0)
        return data

    def allocate_memory(
            self,
            path_im: str, dtype: Union[Type, str] = 'float', data=None,
            shape: tuple = None,
            description: str = 'No description.',
            return_memmap: bool = False, read_mode='r+'):
        axes = self.axes
        axes = axes.replace('C', '') if 'C' in axes else axes
        logger.debug(f'Saving axes as {axes}')
        data = self._ensure_t(data)
        if shape is None:
            shape = self.shape
        if data is None:
            assert shape is not None
            tifffile.imwrite(
                path_im, shape=shape, dtype=dtype, bigtiff=True, metadata={"axes": axes}
            )
        else:
            tifffile.imwrite(
                path_im, data, bigtiff=True, metadata={"axes": axes}
            )
        ome_xml = tifffile.tiffcomment(path_im)
        ome = ome_types.from_xml(ome_xml, parser="lxml")
        ome.images[0].pixels.physical_size_x = self.dim_sizes['X']
        ome.images[0].pixels.physical_size_y = self.dim_sizes['Y']
        ome.images[0].pixels.physical_size_z = self.dim_sizes['Z']
        ome.images[0].pixels.time_increment = self.dim_sizes['T']
        ome.images[0].description = description
        ome.images[0].pixels.type = dtype  # note: numpy uses 8 bits as smallest, so 'bit' type does nothing for bool.
        ome_xml = ome.to_xml()
        tifffile.tiffcomment(path_im, ome_xml)
        if return_memmap:
            return tifffile.memmap(path_im, mode=read_mode)

    def get_im_memmap(self, path_im: str, read_type='r'):
        try:
            im_memmap = tifffile.memmap(path_im, mode=read_type)
        except ValueError:
            if read_type == 'r':
                logger.warning('Could not get memmap, loading file into memory instead.')
                im_memmap = tifffile.imread(path_im)
            else:
                raise ValueError

        if ('C' in self.axes) and (len(im_memmap.shape) == len(self.axes)):
            if self.ch == -1:
                im_memmap = np.max(im_memmap, axis=self.axes.index('C'))
            else:
                im_memmap = np.take(im_memmap, self.ch, axis=self.axes.index('C'))
            # if ch is -1, get a max projection in the C dimension

        self._ensure_t(im_memmap)
        # if len(im_memmap.shape) != len(self.axes) and self.no_t:
        #     # expand dims to match axes
        #     im_memmap = np.expand_dims(im_memmap, axis=self.axes.index('T'))
        return im_memmap

    def create_output_path(self, pipeline_path: str, ext: str = '.ome.tif'):
        if pipeline_path not in self.pipeline_paths:
            self.pipeline_paths[pipeline_path] = os.path.join(self.output_dir,
                                                              f'{self.basename_no_ext}-'
                                                              f'ch{self.ch}-'
                                                              f'{pipeline_path}{ext}')
        return self.pipeline_paths[pipeline_path]


if __name__ == "__main__":
    test_path = "/Users/austin/test_files/nelly_3d/test_no_metadata.tif"
    # test_path = "/Users/austin/test_files/nelly_3d/test.ome.tif"
    im_info = ImInfo(test_path)
    # test_folder = r"D:\test_files\nelly_gav_tests"
    # test_folder = "/Users/austin/test_files/nelly_3d"#/test.ome.tif"
    # all_files = os.listdir(test_folder)
    # all_files = [file for file in all_files if not os.path.isdir(os.path.join(test_folder, file))]
    # im_infos = []
    # for file in all_files:
    #     im_path = os.path.join(test_folder, file)
    #     im_info = ImInfo(im_path)
    #     im_infos.append(im_info)



from qtpy.QtWidgets import QTabWidget

from nellie_napari import NellieProcessor
from nellie_napari.batch_mode import BatchMode
from nellie_napari.home import Home
from nellie_napari.nellie_analysis import NellieAnalysis
from nellie_napari.nellie_fileselect import NellieFileSelect
from nellie_napari.visualizer import NellieVisualizer


class NellieLoader(QTabWidget):
    def __init__(self, napari_viewer: 'napari.viewer.Viewer', parent=None):
        super().__init__(parent)
        self.home = Home(napari_viewer, self)
        self.file_select = NellieFileSelect(napari_viewer, self)
        self.processor = NellieProcessor(napari_viewer, self)
        self.visualizer = NellieVisualizer(napari_viewer, self)
        self.analyzer = NellieAnalysis(napari_viewer, self)
        self.batch_mode = BatchMode(napari_viewer, self)

        self.home_tab = None
        self.file_select_tab = None
        self.processor_tab = None
        self.visualizer_tab = None
        self.analysis_tab = None
        self.batch_tab = None

        self.add_tabs()

        self.im_info = None
        self.valid_files = []

    def add_tabs(self):
        self.home_tab = self.addTab(self.home, "Home")
        self.file_select_tab = self.addTab(self.file_select, "File select")
        self.processor_tab = self.addTab(self.processor, "Process")
        self.visualizer_tab = self.addTab(self.visualizer, "Visualize")
        self.analysis_tab = self.addTab(self.analyzer, "Analyze")
        self.batch_tab = self.addTab(self.batch_mode, "Batch")

        self.setTabEnabled(self.processor_tab, False)
        self.setTabEnabled(self.visualizer_tab, False)
        self.setTabEnabled(self.analysis_tab, False)
        self.setTabEnabled(self.batch_tab, False)

    def reset(self):
        self.setCurrentIndex(self.home_tab)

        # needs to be in reverse order
        self.removeTab(self.batch_tab)
        self.removeTab(self.analysis_tab)
        self.removeTab(self.visualizer_tab)
        self.removeTab(self.processor_tab)
        self.removeTab(self.file_select_tab)

        self.file_select = NellieFileSelect(self.file_select.viewer, self)
        self.processor = NellieProcessor(self.processor.viewer, self)
        self.visualizer = NellieVisualizer(self.visualizer.viewer, self)
        self.analyzer = NellieAnalysis(self.analyzer.viewer, self)
        self.batch_mode = BatchMode(self.batch_mode.viewer, self)

        self.add_tabs()

        self.im_info = None

    def file_ready(self):
        # self.im_info = self.file_select.im_info
        self.processor.post_init()
        self.visualizer.post_init()


if __name__ == "__main__":
    import napari
    viewer = napari.Viewer()
    napari.run()
